from shiny._namespaces import namespace_context, resolve_id


def test_namespaces():
    outer = resolve_id("outer")
    assert outer == "outer"

    with namespace_context(outer):
        # Check if the namespace_context ("outer") is respected during resolve_id
        inner = resolve_id("inner")
        assert inner == "outer-inner"

        # You can also use a ResolvedId as a namespace just by calling it with an id str
        assert outer("inner") == "outer-inner"

        # If an id is already resolved (based on ResolvedId class), resolving it further
        # does nothing
        assert resolve_id(outer) == "outer"

        # When namespace contexts are stacked, inner one wins
        with namespace_context(inner):
            assert resolve_id("inmost") == "outer-inner-inmost"

        # Namespace contexts nest with existing context when string is used
        with namespace_context("inner"):
            assert resolve_id("inmost") == "outer-inner-inmost"

        # Re-installing the same context as is already in place
        with namespace_context(outer):
            assert resolve_id("inmost") == "outer-inmost"

        # You can remove the context with None or ""
        with namespace_context(None):
            assert resolve_id("foo") == "foo"
        with namespace_context(""):
            assert resolve_id("foo") == "foo"

        # Check that this still works after another context was installed/removed
        assert resolve_id("inner") == "outer-inner"



__all__ = (
    "row",
    "column",
    "layout_sidebar",
    "panel_well",
    "panel_sidebar",
    "panel_main",
    "panel_conditional",
    "panel_title",
    "panel_fixed",
    "panel_absolute",
    "help_text",
)

import sys
from typing import Callable, Optional, Union

from shiny.types import MISSING, MISSING_TYPE

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from htmltools import (
    TagChildArg,
    TagAttrArg,
    TagList,
    Tag,
    div,
    tags,
    h2,
    css,
    span,
)

from .._docstring import add_example
from ._html_dependencies import jqui_deps
from ._utils import get_window_title


# TODO: make a python version of the layout guide?
@add_example()
def row(*args: TagChildArg, **kwargs: TagAttrArg) -> Tag:
    """
    Responsive row-column based layout

    Layout UI components using Bootstrap's grid layout system. Use ``row()`` to group
    elements that should appear on the same line (if the browser has adequate width) and
    :func:`~shiny.ui.column` to define how much horizontal space within a 12-unit wide
    grid each on of these elements should occupy. See the `layout guide
    <https://shiny.rstudio.com/articles/layout-guide.html>`_ for more context and
    examples.

    Parameters
    ----------
    args
        Any number of child elements
    kwargs
        Attributes to place on the row tag

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.column`
    """
    return div({"class": "row"}, *args, **kwargs)


def column(
    width: int, *args: TagChildArg, offset: int = 0, **kwargs: TagAttrArg
) -> Tag:
    """
    Responsive row-column based layout

    See :func:`~shiny.ui.row` for more information.

    Parameters
    ----------
    width
        The width of the column (an integer between 1 and 12)
    args
        UI elements to place within the column
    offset
        The number of columns to offset this column from the end of the previous column.
    kwargs
        Attributes to place on the column tag

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.row`
    """

    if width < 1 or width > 12:
        raise ValueError("Column width must be between 1 and 12")
    cls = "col-sm-" + str(width)
    if offset > 0:
        # offset-md-x is for bootstrap 4 forward compat
        # (every size tier has been bumped up one level)
        # https://github.com/twbs/bootstrap/blob/74b8fe7/docs/4.3/migration/index.html#L659
        off = str(offset)
        cls += f" offset-md-{off} col-sm-offset-{off}"
    return div({"class": cls}, *args, **kwargs)


@add_example()
def layout_sidebar(
    # TODO: also accept a generic list (and wrap in panel in that case)?
    sidebar: TagChildArg,
    main: TagChildArg,
    position: Literal["left", "right"] = "left",
) -> Tag:
    """
    Layout a sidebar and main area

    Create a layout with a sidebar (:func:`~shiny.ui.panel_sidebar`) and main area
    (:func:`~shiny.ui.panel_main`). The sidebar is displayed with a distinct background
    color and typically contains input controls. By default, the main area occupies 2/3
    of the horizontal width and typically contains outputs.

    Parameters
    ----------
    sidebar
        A UI element to place in the sidebar (typically a
        :func:`~shiny.ui.panel_sidebar`)
    main
        A UI element to place in the main area (typically a
        :func:`~shiny.ui.panel_main`)
    position
        The position of the sidebar (left or right)

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.panel_sidebar`
    :func:`~shiny.ui.panel_main`
    """
    return row(sidebar, main) if position == "left" else row(main, sidebar)


def panel_well(*args: TagChildArg, **kwargs: TagAttrArg) -> Tag:
    """
    Create a well panel

    Creates a panel with a slightly inset border and grey background. Equivalent to
    Bootstrap's ``well`` CSS class.

    Parameters
    ----------
    args
        UI elements to include inside the panel.
    kwargs
        Attributes to place on the panel tag.

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.panel_sidebar`
    :func:`~shiny.ui.panel_main`
    """
    return div({"class": "well"}, *args, **kwargs)


def panel_sidebar(*args: TagChildArg, width: int = 4, **kwargs: TagAttrArg) -> Tag:
    """
    Create a sidebar panel

    See :func:`~shiny.ui.layout_sidebar` for more information and an example.

    Parameters
    ----------

    args
        UI elements to include inside the sidebar.
    width
        The width of the sidebar (an integer between 1 and 12)
    kwargs
        Attributes to place on the sidebar tag.

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.panel_sidebar`
    :func:`~shiny.ui.panel_main`
    """
    return div(
        {"class": "col-sm-" + str(width)},
        # A11y semantic landmark for sidebar
        tags.form({"class": "well"}, *args, role="complementary", **kwargs),
    )


def panel_main(*args: TagChildArg, width: int = 8, **kwargs: TagAttrArg) -> Tag:
    """
    Create an main area panel

    See :func:`~shiny.ui.layout_sidebar` for more information and an example.

    Parameters
    ----------
    args
        UI elements to include inside the main area.
    width
        The width of the main area (an integer between 1 and 12)
    kwargs
        Attributes to place on the main area tag.

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.panel_sidebar`
    :func:`~shiny.ui.layout_sidebar`
    """
    return div(
        {"class": "col-sm-" + str(width)},
        # A11y semantic landmark for main region
        *args,
        role="main",
        **kwargs,
    )


# TODO: replace `flowLayout()`/`splitLayout()` with a flexbox wrapper?
# def panel_input(*args: TagChild, **kwargs: TagAttr):
#  return div(flowLayout(...), class_="shiny-input-panel")


@add_example()
def panel_conditional(
    condition: str,
    *args: TagChildArg,
    **kwargs: TagAttrArg,
) -> Tag:
    """
    Create a conditional panel

    Show UI elements only if a ``JavaScript`` condition is ``true``.

    Parameters
    ----------
    condition
        A JavaScript expression that will be evaluated repeatedly to determine whether
        the panel should be displayed.
    args
        UI elements to include inside the panel.
    kwargs
        Attributes to place on the panel tag.

    Returns
    -------
    A UI element

    Note
    ----
    In the JS expression, you can refer to input and output JavaScript objects that
    contain the current values of input and output. For example, if you have an input
    with an id of foo, then you can use input.foo to read its value. (Be sure not to
    modify the input/output objects, as this may cause unpredictable behavior.)

    You are not recommended to use special JavaScript characters such as a period . in
    the input id's, but if you do use them anyway, for example, ``id = "foo.bar"``, you
    will have to use ``input["foo.bar"]`` instead of ``input.foo.bar`` to read the input
    value.

    Tip
    ---
    A more powerful (but slower) way to conditionally show UI content is to use
    :func:`~shiny.render.ui`.

    See Also
    -------
    ~shiny.render.ui
    ~shiny.ui.output_ui
    """
    # TODO: do we need a shiny::NS() equivalent?
    ns: Callable[[str], str] = lambda x: x
    return div(*args, data_display_if=condition, data_ns_prefix=ns(""), **kwargs)


@add_example()
def panel_title(
    title: Union[str, Tag, TagList], window_title: Union[str, MISSING_TYPE] = MISSING
) -> TagList:
    """
    Create title(s) for the application.

    Parameters
    ----------
    title
        A title to display in the app's UI.
    window_title
        A title to display on the browser tab.

    Returns
    -------
    A UI element

    Note
    ----
    This result of this function causes a side effect of adding a title tag to the head
    of the document (this is necessary for the browser to display the title in the
    browser window). You can also specify a page title explicitly using the title
    parameter of the top-level page function (e.g., :func:`~shiny.ui.page_fluid`).
    """

    if isinstance(title, str):
        title = h2(title)

    return TagList(get_window_title(title, window_title), title)


def panel_fixed(*args: TagChildArg, **kwargs: TagAttrArg) -> TagList:
    """
    Create a panel of absolutely positioned content.

    This function is equivalent to calling :func:`~shiny.ui.panel_absolute` with
    ``fixed=True`` (i.e., the panel does not scroll with the rest of the page). See
    :func:`~shiny.ui.panel_absolute` for more information.

    Parameters
    ----------
    args
        UI elements to include inside the panel.
    kwargs
        Arguments passed along to :func:`~shiny.ui.panel_absolute`.

    Returns
    -------
    A UI element

    See Also
    -------
    :func:`~shiny.ui.panel_absolute`
    """
    return panel_absolute(*args, fixed=True, **kwargs)


@add_example()
def panel_absolute(
    *args: TagChildArg,
    top: Optional[str] = None,
    left: Optional[str] = None,
    right: Optional[str] = None,
    bottom: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
    draggable: bool = False,
    fixed: bool = False,
    cursor: Literal["auto", "move", "default", "inherit"] = "auto",
    **kwargs: TagAttrArg,
) -> TagList:
    """
    Create a panel of absolutely positioned content.

    Creates a ``<div>`` tag whose CSS position is set to absolute (or fixed if ``fixed =
    True``). The way absolute positioning works in HTML is that absolute coordinates are
    specified relative to its nearest parent element whose position is not set to static
    (which is the default), and if no such parent is found, then relative to the page
    borders. If you're not sure what that means, just keep in mind that you may get
    strange results if you use this function from inside of certain types of panels.

    Parameters
    ----------
    args
        UI elements to include inside the panel.
    top
        Distance between the top of the panel, and the top of the page or parent
        container.
    left
        Distance between the left side of the panel, and the left of the page or parent
        container.
    right
        Distance between the right side of the panel, and the right of the page or
        parent container.
    bottom
        Distance between the bottom of the panel, and the bottom of the page or parent
        container.
    width
        Width of the panel.
    height
        Height of the panel.
    draggable
        If ``True``, allows the user to move the panel by clicking and dragging.
    fixed
        Positions the panel relative to the browser window and prevents it from being
        scrolled with the rest of the page.
    cursor
        The type of cursor that should appear when the user mouses over the panel. Use
        ``"move"`` for a north-east-south-west icon, ``"default"`` for the usual cursor
        arrow, or ``"inherit"`` for the usual cursor behavior (including changing to an
        I-beam when the cursor is over text). The default is ``"auto"``, which is
        equivalent to ``"move" if draggable else "inherit"``.
    kwargs
        Attributes added to the content's container tag.

    Returns
    -------
    A UI element

    Tip
    ----
    The position (``top``, ``left``, ``right``, ``bottom``) and size (``width``,
    ``height``) parameters are all optional, but you should specify exactly two of top,
    bottom, and height and exactly two of left, right, and width for predictable
    results.

    Like most other distance parameters in Shiny, the position and size parameters take
    a number (interpreted as pixels) or a valid CSS size string, such as ``"100px"``
    (100 pixels) or ``"25%"``.

    For arcane HTML reasons, to have the panel fill the page or parent you should
    specify 0 for ``top``, ``left``, ``right``, and ``bottom`` rather than the more
    obvious ``width = "100%"`` and ``height = "100%"``.
    """

    style = css(
        top=top,
        left=left,
        right=right,
        bottom=bottom,
        width=width,
        height=height,
        position="fixed" if fixed else "absolute",
        cursor="move" if draggable else "inherit" if cursor == "auto" else cursor,
    )
    divTag = div(*args, style=style, **kwargs)
    if not draggable:
        return TagList(divTag)
    divTag.add_class("draggable")
    deps = jqui_deps()
    deps.stylesheet = []
    return TagList(deps, divTag, tags.script('$(".draggable").draggable();'))


def help_text(*args: TagChildArg, **kwargs: TagAttrArg) -> Tag:
    """
    Create a help text element

    Parameters
    ----------
    args
        UI elements to include inside the help text.
    kwargs
        Attributes to add to the text container.

    Returns
    -------
    A UI element
    """

    return span({"class": "help-block"}, *args, **kwargs)



from shiny import *
from shiny.types import SilentCancelOutputException

app_ui = ui.page_fluid(
    ui.input_text(
        "txt",
        "Delete the input text completely: it won't get removed below the input",
        "Some text",
        width="400px",
    ),
    ui.output_ui("txt_out"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    def txt_out():
        if not input.txt():
            raise SilentCancelOutputException()
        return "Your input: " + input.txt()


app = App(app_ui, server)



__all__ = (
    "input_select",
    "input_selectize",
)

from typing import List, Mapping, Optional, Tuple, Union, cast

from htmltools import Tag, TagChildArg, TagList, div, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._html_dependencies import selectize_deps
from ._utils import shiny_input_label

_Choices = Mapping[str, TagChildArg]
_OptGrpChoices = Mapping[str, _Choices]

# Canonical format for representing select options.
_SelectChoices = Union[_Choices, _OptGrpChoices]

# Formats available to the user
SelectChoicesArg = Union[
    # ["a", "b", "c"]
    List[str],
    # ("a", "b", "c")
    Tuple[str, ...],
    # {"a": "Choice A", "b": tags.i("Choice B")}
    _Choices,
    # optgroup {"Group A": {"a1": "Choice A1", "a2": tags.i("Choice A2")}, "Group B": {}}
    _OptGrpChoices,
]


_topics = {
    "Server value": """
A list of strings, usually of length 1, with the value of the selected items. When
``multiple=True`` and nothing is selected, this value will be ``None``.
"""
}


@add_example()
def input_selectize(
    id: str,
    label: TagChildArg,
    choices: SelectChoicesArg,
    *,
    selected: Optional[Union[str, List[str]]] = None,
    multiple: bool = False,
    width: Optional[str] = None,
) -> Tag:
    """
    Create a select list that can be used to choose a single or multiple items from a
    list of values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels. A dictionary of dictionaries is
        also supported, and in that case, the top-level keys are treated as
        ``<optgroup>`` labels.
    selected
        The values that should be initially selected, if any.
    multiple
        Is selection of multiple items allowed?
    width
        The CSS width, e.g. '400px', or '100%'

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        A list of strings, usually of length 1, with the value of the selected items. When
        ``multiple=True`` and nothing is selected, this value will be ``None``.

    See Also
    -------
    ~shiny.ui.input_select ~shiny.ui.input_radio_buttons ~shiny.ui.input_checkbox_group
    """

    return input_select(
        id,
        label,
        choices,
        selected=selected,
        multiple=multiple,
        selectize=True,
        width=width,
    )


@add_example()
def input_select(
    id: str,
    label: TagChildArg,
    choices: SelectChoicesArg,
    *,
    selected: Optional[Union[str, List[str]]] = None,
    multiple: bool = False,
    selectize: bool = False,
    width: Optional[str] = None,
    size: Optional[str] = None,
) -> Tag:
    """
    Create a select list that can be used to choose a single or multiple items from a
    list of values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels. A dictionary of dictionaries is
        also supported, and in that case, the top-level keys are treated as
        ``<optgroup>`` labels.
    selected
        The values that should be initially selected, if any.
    multiple
        Is selection of multiple items allowed?
    selectize
        Whether to use selectize.js or not.
    width
        The CSS width, e.g. '400px', or '100%'
    size
        Number of items to show in the selection box; a larger number will result in a
        taller box. Normally, when ``multiple=False``, a select input will be a
        drop-down list, but when size is set, it will be a box instead.

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        A list of strings, usually of length 1, with the value of the selected items. When
        ``multiple=True`` and nothing is selected, this value will be ``None``.

    See Also
    -------
    ~shiny.ui.input_selectize
    ~shiny.ui.update_select
    ~shiny.ui.input_radio_buttons
    ~shiny.ui.input_checkbox_group
    """

    choices_ = _normalize_choices(choices)
    if selected is None:
        selected = _find_first_option(choices_)

    choices_tags = _render_choices(choices_, selected)

    return div(
        shiny_input_label(id, label),
        div(
            tags.select(
                *choices_tags,
                id=resolve_id(id),
                class_=None if selectize else "form-select",
                multiple=multiple,
                width=width,
                size=size,
            ),
            (
                TagList(
                    tags.script("{}", type="application/json", data_for=id),
                    selectize_deps(),
                )
                if selectize
                else None
            ),
        ),
        class_="form-group shiny-input-container",
    )


def _normalize_choices(x: SelectChoicesArg) -> _SelectChoices:
    if x is None:
        raise TypeError("`choices` must be a list, tuple, or dict.")
    elif isinstance(x, (list, tuple)):
        return {k: k for k in x}
    else:
        return x


def _render_choices(
    x: _SelectChoices, selected: Optional[Union[str, List[str]]] = None
) -> TagList:
    result = TagList()

    if x is None:
        return result

    for (k, v) in x.items():
        if isinstance(v, Mapping):
            result.append(
                tags.optgroup(
                    *(_render_choices(cast(_SelectChoices, v), selected)), label=k
                )
            )
        else:
            is_selected = False
            if isinstance(selected, list):
                is_selected = k in selected
            else:
                is_selected = k == selected

            result.append(tags.option(v, value=k, selected=is_selected))

    return result


# Returns the first option in a _SelectChoices object. For most cases, this is
# straightforward. In the following, the first option is "a":
# {"a": "Choice A", "b": "Choice B", "c": "Choice C"}
#
# Sometimes the first option is nested within an optgroup. For example, in the
# following, the first option is "b1":
# {
#     "Group A": {},
#     "Group B": {"Choice B1": "b1", "Choice B2": "b2"},
# }
def _find_first_option(x: _SelectChoices) -> Optional[str]:
    if x is None:
        return None

    for (k, v) in x.items():
        if isinstance(v, dict):
            result = _find_first_option(cast(_SelectChoices, v))
            if result is not None:
                return result
        else:
            return k

    return None



from shiny import *

app_ui = ui.page_fluid(
    ui.input_text("msg", "Enter a message"),
    ui.input_action_button("submit", "Submit the message"),
    # It'd be better to use ui.insert_ui() in order to implement this kind of
    # functionality...this is just a basic demo of how custom message handling works.
    ui.tags.div(id="messages"),
    ui.tags.script(
        """
        $(function() {
            Shiny.addCustomMessageHandler("append_msg", function(message) {
                $("<p>").text(message.msg).appendTo("#messages");
            });
        });
        """
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    @reactive.event(input.submit)
    async def _():
        await session.send_custom_message("append_msg", {"msg": input.msg()})


app = App(app_ui, server, debug=True)



from shiny import *
from shiny.types import SilentException

app_ui = ui.page_fluid(
    ui.input_text(
        "txt",
        "Enter text to see it displayed below the input",
        width="400px",
    ),
    ui.output_ui("txt_out"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    def txt_out():
        if not input.txt():
            raise SilentException()
        return "Your input: " + input.txt()


app = App(app_ui, server)



import logging
import os
import re
from ipaddress import ip_address
from subprocess import run
import typing
from typing import Dict, Pattern
from urllib.parse import urlparse


def is_workbench() -> bool:
    return bool(os.getenv("RS_SERVER_URL") and os.getenv("RS_SESSION_URL"))


def is_proxy_env() -> bool:
    return is_workbench()


port_cache: Dict[int, str] = {}


def get_proxy_url(url: str) -> str:
    if not is_workbench():
        return url

    parts = urlparse(url)
    is_loopback = parts.hostname == "localhost"
    if not is_loopback:
        try:
            is_loopback = ip_address(parts.hostname or "").is_loopback
        except ValueError:
            pass
    if not is_loopback:
        return url

    path = parts.path or "/"

    server_url = os.getenv("RS_SERVER_URL", "")
    session_url = os.getenv("RS_SESSION_URL", "")

    if parts.scheme.lower() in ["ws", "wss"]:
        server_url = re.sub("^http", "ws", server_url)
    server_url = re.sub("/$", "", server_url)
    session_url = re.sub("^/", "", session_url)

    port = (
        parts.port
        if parts.port
        else 80
        if parts.scheme in ["ws", "http"]
        else 443
        if parts.scheme in ["wss", "https"]
        else 0
    )
    if port == 0:
        return url

    if port in port_cache:
        ptoken = port_cache[port]
    else:
        try:
            res = run(
                ["/usr/lib/rstudio-server/bin/rserver-url", str(port)],
                capture_output=True,
                encoding="ascii",
            )
        except FileNotFoundError:
            return url
        if res.returncode != 0:
            return url
        ptoken = res.stdout
        port_cache[port] = ptoken

    return f"{server_url}/{session_url}p/{ptoken}{path}{'?' if parts.query else ''}{parts.query}"


pat_local_url: Pattern[str] = re.compile(
    "https?://(127.0.0.1|localhost)(:\\d+)?([-\\+=&;%@.\\w_]*)", re.IGNORECASE
)


class ProxyUrlFilter:
    def __init__(self):
        pass

    def filter(self, record: logging.LogRecord) -> int:
        record.msg = pat_local_url.sub(self.url_replacement, record.getMessage())
        if hasattr(record, "color_message"):
            color_msg = str(record.color_message)  # type: ignore
            if record.args:
                color_msg = color_msg % record.args
            record.color_message = pat_local_url.sub(  # type: ignore
                self.url_replacement, color_msg
            )
        record.args = ()
        return 1

    def url_replacement(self, match: typing.Match[str]) -> str:
        return get_proxy_url(match.group(0))



"""Tests for `Module`."""

import asyncio
from typing import Dict, Union, cast

import pytest

from htmltools import Tag, TagList
from shiny import *
from shiny._connection import MockConnection
from shiny._namespaces import resolve_id
from shiny.session import get_current_session


@module.ui
def mod_inner_ui() -> TagList:
    return TagList(
        ui.input_action_button("button", label="inner"),
        ui.output_text(resolve_id("out")),
    )


@module.ui
def mod_outer_ui() -> TagList:
    return TagList(mod_inner_ui("inner"), ui.output_text("out2"))


def get_id(x: TagList, child_idx: int = 0) -> str:
    return cast(Tag, x[child_idx]).attrs["id"]


def test_module_ui():
    x = mod_inner_ui("inner")
    assert get_id(x, 0) == "inner-button"
    assert get_id(x, 1) == "inner-out"
    y = mod_outer_ui("outer")
    assert get_id(y, 0) == "outer-inner-button"
    assert get_id(y, 1) == "outer-inner-out"
    assert get_id(y, 2) == "outer-out2"


@pytest.mark.asyncio
async def test_session_scoping():

    sessions: Dict[str, Union[Session, None, str]] = {}

    @module.server
    def inner_server(input: Inputs, output: Outputs, session: Session):
        @reactive.Calc
        def out():
            return get_current_session()

        @reactive.Effect
        def _():
            sessions["inner"] = session
            sessions["inner_current"] = get_current_session()
            sessions["inner_calc_current"] = out()
            sessions["inner_id"] = session.ns("foo")
            sessions["inner_ui_id"] = get_id(mod_outer_ui("outer"), 0)

    @module.server
    def outer_server(input: Inputs, output: Outputs, session: Session):
        @reactive.Calc
        def out():
            return get_current_session()

        @reactive.Effect
        def _():
            inner_server("mod_inner")
            sessions["outer"] = session
            sessions["outer_current"] = get_current_session()
            sessions["outer_calc_current"] = out()
            sessions["outer_id"] = session.ns("foo")
            sessions["outer_ui_id"] = get_id(mod_outer_ui("outer"), 0)

    def server(input: Inputs, output: Outputs, session: Session):
        outer_server("mod_outer")

        @reactive.Calc
        def out():
            return get_current_session()

        @reactive.Effect
        def _():
            sessions["top"] = session
            sessions["top_current"] = get_current_session()
            sessions["top_calc_current"] = out()
            sessions["top_id"] = session.ns("foo")
            sessions["top_ui_id"] = get_id(mod_outer_ui("outer"), 0)

    conn = MockConnection()
    sess = App(ui.TagList(), server)._create_session(conn)

    async def mock_client():
        conn.cause_receive('{"method":"init","data":{}}')
        conn.cause_disconnect()

    await asyncio.gather(mock_client(), sess._run())

    assert sessions["inner"] is sessions["inner_current"]
    assert sessions["inner_current"] is sessions["inner_calc_current"]
    assert isinstance(sessions["inner_current"], Session)
    assert sessions["inner_id"] == "mod_outer-mod_inner-foo"
    assert sessions["inner_ui_id"] == "mod_outer-mod_inner-outer-inner-button"

    assert sessions["outer"] is sessions["outer_current"]
    assert sessions["outer_current"] is sessions["outer_calc_current"]
    assert isinstance(sessions["outer_current"], Session)
    assert sessions["outer_id"] == "mod_outer-foo"
    assert sessions["outer_ui_id"] == "mod_outer-outer-inner-button"

    assert sessions["top"] is sessions["top_current"]
    assert sessions["top_current"] is sessions["top_calc_current"]
    assert isinstance(sessions["top_current"], Session)
    assert sessions["top_id"] == "foo"
    assert sessions["top_ui_id"] == "outer-inner-button"



from shiny import *

app_ui = ui.page_fluid(
    ui.input_action_button("add", "Add more controls"),
    ui.output_ui("moreControls"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.ui
    @reactive.event(input.add)
    def moreControls():
        return ui.TagList(
            ui.input_slider("n", "N", min=1, max=1000, value=500),
            ui.input_text("label", "Label"),
        )


app = App(app_ui, server)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

from sphinx.application import Sphinx

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("../sphinxext"))

# -- Project information -----------------------------------------------------

project = "Shiny"
copyright = "2022, RStudio"
author = "RStudio"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "myst_parser",
    "pyshinyapp",  # custom shiny extention for embedded apps in docs
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# https://sphinx-book-theme.readthedocs.io/en/latest/tutorials/get-started.html
html_theme = "sphinx_book_theme"

html_theme_options = {
    "repository_url": "https://github.com/rstudio/py-shiny",
    "use_repository_button": True,
}

html_title = "for Python"

html_logo = "images/shiny-logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Shiny specific options --------------------------------------------------

# Tell @shiny._docstring.add_example() to actually add examples to the docstrings.
os.environ["SHINY_ADD_EXAMPLES"] = "true"


# WARNING: if you're thinking about adding subclasses of docutils.nodes.Element here,
# don't! You'll get unhelpful pickling errors. https://github.com/sphinx-doc/sphinx/pull/6754
# Instead, add them to the docs/source/sphinxext/ directory and add the name to the
# extensions list above (as done for pyshinyapp).


def setup(app: Sphinx) -> None:
    app.add_js_file("js/fix-logo-link.js")
    app.add_js_file("js/disable-keypress.js")



from shiny import *

app_ui = ui.page_fluid(
    ui.input_numeric("obs", "Observations:", 10, min=1, max=100),
    ui.output_text_verbatim("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.text
    def value():
        return input.obs()


app = App(app_ui, server)



__all__ = (
    "RenderFunction",
    "RenderFunctionAsync",
    "RenderText",
    "RenderTextAsync",
    "text",
    "RenderPlot",
    "RenderPlotAsync",
    "plot",
    "RenderImage",
    "RenderImageAsync",
    "image",
    "RenderTable",
    "RenderTableAsync",
    "table",
    "RenderUI",
    "RenderUIAsync",
    "ui",
)

import base64
import os
import sys
import typing
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
    overload,
)

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

from htmltools import TagChildArg

# These aren't  used directly in this file, but they seems necessary for Sphinx to work
# cleanly.
from htmltools import (  # pyright: ignore[reportUnusedImport] # noqa: F401
    Tagifiable,  # pyright: ignore[reportUnusedImport]
    Tag,  # pyright: ignore[reportUnusedImport]
    TagList,  # pyright: ignore[reportUnusedImport]
)

if TYPE_CHECKING:
    from ..session import Session
    from ..session._utils import RenderedDeps

from .. import _utils
from ..types import ImgData
from ._try_render_plot import (
    TryPlotResult,
    try_render_matplotlib,
    try_render_pil,
    try_render_plotnine,
)

# Input type for the user-spplied function that is passed to a render.xx
IT = TypeVar("IT")
# Output type after the RenderFunction.__call__ method is called on the IT object.
OT = TypeVar("OT")


# ======================================================================================
# RenderFunction/RenderFunctionAsync base class
# ======================================================================================

# A RenderFunction object is given a user-provided function which returns an IT. When
# the .__call___ method is invoked, it calls the user-provided function (which returns
# an IT), then converts the IT to an OT. Note that in many cases but not all, IT and OT
# will be the same.
class RenderFunction(Generic[IT, OT]):
    def __init__(self, fn: Callable[[], IT]) -> None:
        self.__name__ = fn.__name__
        self.__doc__ = fn.__doc__

    def __call__(self) -> OT:
        raise NotImplementedError

    def set_metadata(self, session: "Session", name: str) -> None:
        """When RenderFunctions are assigned to Output object slots, this method
        is used to pass along session and name information.
        """
        self._session: Session = session
        self._name: str = name


# The reason for having a separate RenderFunctionAsync class is because the __call__
# method is marked here as async; you can't have a single class where one method could
# be either sync or async.
class RenderFunctionAsync(RenderFunction[IT, OT]):
    async def __call__(self) -> OT:  # type: ignore
        raise NotImplementedError


# ======================================================================================
# RenderText
# ======================================================================================
RenderTextFunc = Callable[[], Union[str, None]]
RenderTextFuncAsync = Callable[[], Awaitable[Union[str, None]]]


class RenderText(RenderFunction[Union[str, None], Union[str, None]]):
    def __init__(self, fn: RenderTextFunc) -> None:
        super().__init__(fn)
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderTextFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> Union[str, None]:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> Union[str, None]:
        res = await self._fn()
        if res is None:
            return None
        return str(res)


class RenderTextAsync(
    RenderText, RenderFunctionAsync[Union[str, None], Union[str, None]]
):
    def __init__(self, fn: RenderTextFuncAsync) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderTextFunc, fn))

    async def __call__(self) -> Union[str, None]:  # type: ignore
        return await self._run()


@overload
def text(fn: Union[RenderTextFunc, RenderTextFuncAsync]) -> RenderText:
    ...


@overload
def text() -> Callable[[Union[RenderTextFunc, RenderTextFuncAsync]], RenderText]:
    ...


def text(
    fn: Optional[Union[RenderTextFunc, RenderTextFuncAsync]] = None
) -> Union[
    RenderText, Callable[[Union[RenderTextFunc, RenderTextFuncAsync]], RenderText]
]:
    """
    Reactively render text.

    Returns
    -------
    A decorator for a function that returns a string.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_text` container (see :func:`~shiny.ui.output_text` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_text
    """

    def wrapper(fn: Union[RenderTextFunc, RenderTextFuncAsync]) -> RenderText:
        if _utils.is_async_callable(fn):
            return RenderTextAsync(fn)
        else:
            fn = typing.cast(RenderTextFunc, fn)
            return RenderText(fn)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderPlot
# ======================================================================================
# It would be nice to specify the return type of RenderPlotFunc to be something like:
#   Union[matplotlib.figure.Figure, PIL.Image.Image]
# However, if we did that, we'd have to import those modules at load time, which adds
# a nontrivial amount of overhead. So for now, we're just using `object`.
RenderPlotFunc = Callable[[], object]
RenderPlotFuncAsync = Callable[[], Awaitable[object]]


class RenderPlot(RenderFunction[object, Union[ImgData, None]]):
    _ppi: float = 96

    def __init__(
        self, fn: RenderPlotFunc, *, alt: Optional[str] = None, **kwargs: object
    ) -> None:
        super().__init__(fn)
        self._alt: Optional[str] = alt
        self._kwargs = kwargs
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderPlotFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> Union[ImgData, None]:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> Union[ImgData, None]:
        # Reactively read some information about the plot.
        pixelratio: float = typing.cast(
            float, self._session.input[".clientdata_pixelratio"]()
        )
        width: float = typing.cast(
            float, self._session.input[f".clientdata_output_{self._name}_width"]()
        )
        height: float = typing.cast(
            float, self._session.input[f".clientdata_output_{self._name}_height"]()
        )

        x = await self._fn()

        if x is None:
            return None

        # Try each type of renderer in turn. The reason we do it this way is to avoid
        # importing modules that aren't already loaded. That could slow things down, or
        # worse, cause an error if the module isn't installed.
        #
        # Each try_render function should return either an ImgResult, None (which
        # indicates that the rendering failed), or the string "TYPE_MISMATCH" (which
        # indicate that `fig` object was not the type of object that the renderer knows
        # how to handle). In the case of a "TYPE_MISMATCH", it will move on to the next
        # renderer.
        result: TryPlotResult = None

        if "plotnine" in sys.modules:
            result = try_render_plotnine(
                x, width, height, pixelratio, self._ppi, **self._kwargs
            )
            if result != "TYPE_MISMATCH":
                return result

        if "matplotlib" in sys.modules:
            result = try_render_matplotlib(
                x, width, height, pixelratio, self._ppi, **self._kwargs
            )
            if result != "TYPE_MISMATCH":
                return result

        if "PIL" in sys.modules:
            result = try_render_pil(
                x, width, height, pixelratio, self._ppi, **self._kwargs
            )
            if result != "TYPE_MISMATCH":
                return result

        raise Exception(
            f"@render.plot doesn't know to render objects of type '{str(type(x))}'. "
            + "Consider either requesting support for this type of plot object, and/or "
            + " explictly saving the object to a (png) file and using @render.image."
        )


class RenderPlotAsync(RenderPlot, RenderFunctionAsync[object, Union[ImgData, None]]):
    def __init__(
        self, fn: RenderPlotFuncAsync, alt: Optional[str] = None, **kwargs: Any
    ) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderPlotFunc, fn), alt=alt, **kwargs)

    async def __call__(self) -> Union[ImgData, None]:  # type: ignore
        return await self._run()


@overload
def plot(fn: Union[RenderPlotFunc, RenderPlotFuncAsync]) -> RenderPlot:
    ...


@overload
def plot(
    *,
    alt: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[Union[RenderPlotFunc, RenderPlotFuncAsync]], RenderPlot]:
    ...


# TODO: Use more specific types for render.plot
def plot(
    fn: Optional[Union[RenderPlotFunc, RenderPlotFuncAsync]] = None,
    *,
    alt: Optional[str] = None,
    **kwargs: Any,
) -> Union[
    RenderPlot, Callable[[Union[RenderPlotFunc, RenderPlotFuncAsync]], RenderPlot]
]:
    """
    Reactively render a plot object as an HTML image.

    Parameters
    ----------
    alt
        Alternative text for the image if it cannot be displayed or viewed (i.e., the
        user uses a screen reader).
    **kwargs
        Additional keyword arguments passed to the relevant method for saving the image
        (e.g., for matplotlib, arguments to ``savefig()``; for PIL and plotnine,
        arguments to ``save()``).

    Returns
    -------
    A decorator for a function that returns any of the following:

        1. A :class:`matplotlib.figure.Figure` instance.
        2. An :class:`matplotlib.artist.Artist` instance.
        3. A list/tuple of Figure/Artist instances.
        4. An object with a 'figure' attribute pointing to a
           :class:`matplotlib.figure.Figure` instance.
        5. A :class:`PIL.Image.Image` instance.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of a
    :func:`~shiny.ui.output_plot` container (see :func:`~shiny.ui.output_plot` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_plot
    ~shiny.render.image
    """

    def wrapper(fn: Union[RenderPlotFunc, RenderPlotFuncAsync]) -> RenderPlot:
        if _utils.is_async_callable(fn):
            return RenderPlotAsync(fn, alt=alt, **kwargs)
        else:
            return RenderPlot(fn, alt=alt, **kwargs)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderImage
# ======================================================================================
RenderImageFunc = Callable[[], Union[ImgData, None]]
RenderImageFuncAsync = Callable[[], Awaitable[Union[ImgData, None]]]


class RenderImage(RenderFunction[Union[ImgData, None], Union[ImgData, None]]):
    def __init__(
        self,
        fn: RenderImageFunc,
        *,
        delete_file: bool = False,
    ) -> None:
        super().__init__(fn)
        self._delete_file: bool = delete_file
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderImageFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> Union[ImgData, None]:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> Union[ImgData, None]:
        res: Union[ImgData, None] = await self._fn()
        if res is None:
            return None

        src: str = res.get("src")
        try:
            with open(src, "rb") as f:
                data = base64.b64encode(f.read())
                data_str = data.decode("utf-8")
            content_type = _utils.guess_mime_type(src)
            res["src"] = f"data:{content_type};base64,{data_str}"
            return res
        finally:
            if self._delete_file:
                os.remove(src)


class RenderImageAsync(
    RenderImage, RenderFunctionAsync[Union[ImgData, None], Union[ImgData, None]]
):
    def __init__(self, fn: RenderImageFuncAsync, delete_file: bool = False) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderImageFunc, fn), delete_file=delete_file)

    async def __call__(self) -> Union[ImgData, None]:  # type: ignore
        return await self._run()


@overload
def image(fn: Union[RenderImageFunc, RenderImageFuncAsync]) -> RenderImage:
    ...


@overload
def image(
    *,
    delete_file: bool = False,
) -> Callable[[Union[RenderImageFunc, RenderImageFuncAsync]], RenderImage]:
    ...


def image(
    fn: Optional[Union[RenderImageFunc, RenderImageFuncAsync]] = None,
    *,
    delete_file: bool = False,
) -> Union[
    RenderImage, Callable[[Union[RenderImageFunc, RenderImageFuncAsync]], RenderImage]
]:
    """
    Reactively render a image file as an HTML image.

    Parameters
    ----------
    delete_file
        If ``True``, the image file will be deleted after rendering.

    Returns
    -------
    A decorator for a function that returns an `~shiny.types.ImgData` object.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_image` container (see :func:`~shiny.ui.output_image` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_image
    ~shiny.types.ImgData
    ~shiny.render.plot
    """

    def wrapper(fn: Union[RenderImageFunc, RenderImageFuncAsync]) -> RenderImage:
        if _utils.is_async_callable(fn):
            return RenderImageAsync(fn, delete_file=delete_file)
        else:
            fn = typing.cast(RenderImageFunc, fn)
            return RenderImage(fn, delete_file=delete_file)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderTable
# ======================================================================================
# It would be nice to specify the return type of RenderPlotFunc to be something like:
#   Union[pandas.DataFrame, <protocol with .to_pandas()>]
# However, if we did that, we'd have to import pandas at load time, which adds
# a nontrivial amount of overhead. So for now, we're just using `object`.
RenderTableFunc = Callable[[], object]
RenderTableFuncAsync = Callable[[], Awaitable[object]]


@runtime_checkable
class PandasCompatible(Protocol):
    # Signature doesn't matter, runtime_checkable won't look at it anyway
    def to_pandas(self) -> object:
        ...


class RenderTable(RenderFunction[object, Union["RenderedDeps", None]]):
    def __init__(
        self,
        fn: RenderTableFunc,
        *,
        index: bool = False,
        classes: str = "table shiny-table w-auto",
        border: int = 0,
        **kwargs: object,
    ) -> None:
        super().__init__(fn)
        self._index = index
        self._classes = classes
        self._border = border
        self._kwargs = kwargs
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderTableFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> Union["RenderedDeps", None]:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> Union["RenderedDeps", None]:
        x = await self._fn()

        if x is None:
            return None

        import pandas  # pyright: reportMissingTypeStubs=false,reportUnknownVariableType=false,reportMissingImports=false,reportMissingModuleSource=false
        import pandas.io.formats.style

        html: str
        if isinstance(
            x, pandas.io.formats.style.Styler
        ):  # pyright: reportUnknownMemberType=false
            html = x.to_html(**self._kwargs)  # pyright: reportGeneralTypeIssues=false
        else:
            if not isinstance(x, pandas.DataFrame):
                if not isinstance(x, PandasCompatible):
                    raise TypeError(
                        "@render.table doesn't know how to render objects of type "
                        f"'{str(type(x))}'. Return either a pandas.DataFrame, or an object "
                        "that has a .to_pandas() method."
                    )
                x = x.to_pandas()

            df = typing.cast(pandas.DataFrame, x)
            html = df.to_html(
                index=self._index,
                classes=self._classes,
                border=self._border,
                **self._kwargs,
            )
        return {"deps": [], "html": html}


class RenderTableAsync(RenderTable, RenderFunctionAsync[object, Union[ImgData, None]]):
    def __init__(
        self,
        fn: RenderTableFuncAsync,
        *,
        index: bool = False,
        classes: str = "table shiny-table w-auto",
        border: int = 0,
        **kwargs: Any,
    ) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(
            typing.cast(RenderTableFunc, fn),
            index=index,
            classes=classes,
            border=border,
            **kwargs,
        )

    async def __call__(self) -> Union["RenderedDeps", None]:  # type: ignore
        return await self._run()


@overload
def table(fn: Union[RenderTableFunc, RenderTableFuncAsync]) -> RenderTable:
    ...


@overload
def table(
    *,
    index: bool = False,
    classes: str = "table shiny-table w-auto",
    border: int = 0,
    **kwargs: Any,
) -> Callable[[Union[RenderTableFunc, RenderTableFuncAsync]], RenderTable]:
    ...


# TODO: Use more specific types for render.table
def table(
    fn: Optional[Union[RenderTableFunc, RenderTableFuncAsync]] = None,
    *,
    index: bool = False,
    classes: str = "table shiny-table w-auto",
    border: int = 0,
    **kwargs: Any,
) -> Union[
    RenderTable, Callable[[Union[RenderTableFunc, RenderTableFuncAsync]], RenderTable]
]:
    """
    Reactively render a Pandas data frame object (or similar) as a basic HTML table.

    Parameters
    ----------
    index
        Whether to print index (row) labels. (Ignored for pandas :class:`Styler`
        objects; call ``style.hide(axis="index")`` from user code instead.)
    classes
        CSS classes (space separated) to apply to the resulting table. By default, we
        use `table shiny-table w-auto` which is designed to look reasonable with Bootstrap 5.
        (Ignored for pandas :class:`Styler` objects; call
        ``style.set_table_attributes('class="dataframe table shiny-table w-auto"')``
        from user code instead.)
    **kwargs
        Additional keyword arguments passed to ``pandas.DataFrame.to_html()`` or
        ``pandas.io.formats.style.Styler.to_html()``.

    Returns
    -------
    A decorator for a function that returns any of the following:

        1. A pandas :class:`DataFrame` object.
        2. A pandas :class:`Styler` object.
        3. Any object that has a `.to_pandas()` method (e.g., a Polars data frame or
           Arrow table).

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_table` container (see :func:`~shiny.ui.output_table` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_table
    """

    def wrapper(fn: Union[RenderTableFunc, RenderTableFuncAsync]) -> RenderTable:
        if _utils.is_async_callable(fn):
            return RenderTableAsync(
                fn, index=index, classes=classes, border=border, **kwargs
            )
        else:
            return RenderTable(
                fn, index=index, classes=classes, border=border, **kwargs
            )

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderUI
# ======================================================================================
RenderUIFunc = Callable[[], TagChildArg]
RenderUIFuncAsync = Callable[[], Awaitable[TagChildArg]]


class RenderUI(RenderFunction[TagChildArg, Union["RenderedDeps", None]]):
    def __init__(self, fn: RenderUIFunc) -> None:
        super().__init__(fn)
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderUIFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> Union["RenderedDeps", None]:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> Union["RenderedDeps", None]:
        ui: TagChildArg = await self._fn()
        if ui is None:
            return None

        return self._session._process_ui(ui)


class RenderUIAsync(
    RenderUI, RenderFunctionAsync[TagChildArg, Union["RenderedDeps", None]]
):
    def __init__(self, fn: RenderUIFuncAsync) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderUIFunc, fn))

    async def __call__(self) -> Union["RenderedDeps", None]:  # type: ignore
        return await self._run()


@overload
def ui(fn: Union[RenderUIFunc, RenderUIFuncAsync]) -> RenderUI:
    ...


@overload
def ui() -> Callable[[Union[RenderUIFunc, RenderUIFuncAsync]], RenderUI]:
    ...


def ui(
    fn: Optional[Union[RenderUIFunc, RenderUIFuncAsync]] = None
) -> Union[RenderUI, Callable[[Union[RenderUIFunc, RenderUIFuncAsync]], RenderUI]]:
    """
    Reactively render HTML content.

    Returns
    -------
    A decorator for a function that returns an object of type `~shiny.ui.TagChildArg`.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_ui` container (see :func:`~shiny.ui.output_ui` for example
    usage).

    See Also
    --------
    ~shiny.ui.output_ui
    """

    def wrapper(
        fn: Union[Callable[[], TagChildArg], Callable[[], Awaitable[TagChildArg]]]
    ) -> RenderUI:
        if _utils.is_async_callable(fn):
            return RenderUIAsync(fn)
        else:
            fn = typing.cast(RenderUIFunc, fn)
            return RenderUI(fn)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)



__all__ = ("input_handlers",)

from datetime import date, datetime
from typing import TYPE_CHECKING, Callable, Dict, Union, List, Any, Tuple

if TYPE_CHECKING:
    from .session import Session

from .types import ActionButtonValue

InputHandlerType = Callable[[Any, str, "Session"], Any]


class _InputHandlers(Dict[str, InputHandlerType]):
    def __init__(self):
        super().__init__()

    def add(self, type: str, force: bool = False) -> Callable[[InputHandlerType], None]:
        def _(func: InputHandlerType):
            if type in self and not force:
                raise ValueError(f"Input handler {type} already registered")
            self[type] = func
            return None

        return _

    def remove(self, type: str):
        del self[type]

    def _process_value(
        self, type: str, value: Any, name: str, session: "Session"
    ) -> Any:
        handler = self.get(type)
        if handler is None:
            raise ValueError("No input handler registered for type: " + type)
        return handler(value, name, session)


input_handlers: _InputHandlers = _InputHandlers()
input_handlers.__doc__ = """
Manage Shiny input handlers.

Add and/or remove input handlers of a given ``type``. Shiny uses these handlers to
pre-process input values from the client (after being deserialized) before passing them
to the ``input`` argument of an :func:`~shiny.App`'s ``server`` function.

The ``type`` is based on the ``getType()`` JavaScript method on the relevant Shiny
input binding. See `this article <https://shiny.rstudio.com/articles/js-custom-input.html>`_
for more information on how to create custom input bindings.

Methods
--------
add(type: str, force: bool = False) -> Callable[[InputHandlerType], None]
    Register an input handler. This method returns a decorator that registers the
    decorated function as the handler for the given ``type``. This handler should
    accept three arguments:
    - the input ``value``
    - the input ``name``
    - the :class:`~shiny.Session` object
remove(type: str)
    Unregister an input handler.

Note
----
``add()`` ing an input handler will make it persist for the duration of the Python
process (unless Shiny is explicitly reloaded). For that reason, verbose naming is
encouraged to minimize the risk of colliding with other Shiny input binding(s) which
happen to use the same ``type`` (if this the binding is bundled with a package, we
recommend the format of "packageName.widgetName").

Example
-------
.. code-block:: python

    from shiny.input_handler import input_handlers
    @input_handlers.add("mypackage.intify")
    def _(value, name, session):
        return int(value)

On the Javascript side, the associated input binding must have a corresponding
``getType`` method:

.. code-block:: javascript

    getType: function(el) {
      return "mypackage.intify";
    }
"""


@input_handlers.add("shiny.date")
def _(
    value: Union[str, List[str]], name: str, session: "Session"
) -> Union[date, Tuple[date, date]]:
    if isinstance(value, str):
        return datetime.strptime(value, "%Y-%m-%d").date()
    return tuple(datetime.strptime(v, "%Y-%m-%d").date() for v in value)


@input_handlers.add("shiny.datetime")
def _(
    value: Union[int, float, List[int], List[float]], name: str, session: "Session"
) -> Union[datetime, Tuple[datetime, datetime]]:
    if isinstance(value, (int, float)):
        return datetime.utcfromtimestamp(value)
    return tuple(datetime.utcfromtimestamp(v) for v in value)


@input_handlers.add("shiny.action")
def _(value: int, name: str, session: "Session") -> ActionButtonValue:
    # TODO: ActionButtonValue() class can probably be removed
    return ActionButtonValue(value)


# The inputs handlers below currently do nothing, but still need to be defined,
# otherwise there will be an error when the input value is handled.


@input_handlers.add("shiny.number")
def _(value: str, name: str, session: "Session") -> str:
    return value


# TODO: implement when we have bookmarking
@input_handlers.add("shiny.password")
def _(value: str, name: str, session: "Session") -> str:
    return value


# TODO: implement when we have bookmarking
@input_handlers.add("shiny.file")
def _(value: Any, name: str, session: "Session") -> Any:
    return value



from shiny import *

# ============================================================
# Counter module
# ============================================================
@module.ui
def counter_ui(label: str = "Increment counter") -> ui.TagChildArg:
    return ui.div(
        {"style": "border: 1px solid #ccc; border-radius: 5px; margin: 5px 0;"},
        ui.h2("This is " + label),
        ui.input_action_button(id="button", label=label),
        ui.output_text_verbatim(id="out"),
    )


@module.server
def counter_server(
    input: Inputs, output: Outputs, session: Session, starting_value: int = 0
):
    count: reactive.Value[int] = reactive.Value(starting_value)

    @reactive.Effect
    @reactive.event(input.button)
    def _():
        count.set(count() + 1)

    @output
    @render.text
    def out() -> str:
        return f"Click count is {count()}"


# =============================================================================
# App that uses module
# =============================================================================
app_ui = ui.page_fluid(
    counter_ui("counter1", "Counter 1"),
    counter_ui("counter2", "Counter 2"),
)


def server(input: Inputs, output: Outputs, session: Session):
    counter_server("counter1")
    counter_server("counter2")


app = App(app_ui, server)



import copy
import importlib
import importlib.util
import os
import re
import shutil
import sys
import types
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import click
import uvicorn
import uvicorn.config

import shiny

from . import _autoreload, _hostenv, _static


@click.group()  # pyright: ignore[reportUnknownMemberType]
def main() -> None:
    pass


stop_shortcut = "Ctrl+C"


@main.command(
    help=f"""Runs a Shiny app. Press {stop_shortcut} to stop.

The APP argument indicates the Python file (or module) and attribute where the
shiny.App object can be found. For example, if the current directory contains
a file called app.py, which contains an `app` variable that is the Shiny app,
any of the following will work:

\b
  * shiny run             # app:app is assumed
  * shiny run app         # :app is assumed
  * shiny run app.py      # :app is assumed
  * shiny run app:app
  * shiny run app.py:app
"""
)
@click.argument("app", default="app.py:app")
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=8000,
    help="Bind socket to this port.",
    show_default=True,
)
@click.option(
    "--autoreload-port",
    type=str,
    default="+123",
    help="Bind autoreload socket to this port number. If the value begins with + or -, it will be added to the value of --port. Ignored if --reload is not used.",
    show_default=True,
)
@click.option(
    "--debug", is_flag=True, default=False, help="Enable debug mode.", hidden=True
)
@click.option("--reload", is_flag=True, default=False, help="Enable auto-reload.")
@click.option(
    "--ws-max-size",
    type=int,
    default=16777216,
    help="WebSocket max size message in bytes",
    show_default=True,
)
@click.option(
    "--log-level",
    type=click.Choice(list(uvicorn.config.LOG_LEVELS.keys())),
    default=None,
    help="Log level. [default: info]",
    show_default=True,
)
@click.option(
    "--app-dir",
    default=".",
    show_default=True,
    help="Look for APP in the specified directory, by adding this to the PYTHONPATH."
    " Defaults to the current working directory. If APP is a file path, this argument"
    " is ignored.",
)
@click.option(
    "--factory",
    is_flag=True,
    default=False,
    help="Treat APP as an application factory, i.e. a () -> <ASGI app> callable.",
    show_default=True,
)
def run(
    app: Union[str, shiny.App],
    host: str,
    port: int,
    autoreload_port: str,
    debug: bool,
    reload: bool,
    ws_max_size: int,
    log_level: str,
    app_dir: str,
    factory: bool,
) -> None:
    return run_app(
        app,
        host=host,
        port=port,
        autoreload_port=autoreload_port,
        debug=debug,
        reload=reload,
        ws_max_size=ws_max_size,
        log_level=log_level,
        app_dir=app_dir,
        factory=factory,
    )


def run_app(
    app: Union[str, shiny.App] = "app:app",
    host: str = "127.0.0.1",
    port: int = 8000,
    autoreload_port: str = "",
    debug: bool = False,
    reload: bool = False,
    ws_max_size: int = 16777216,
    log_level: Optional[str] = None,
    app_dir: Optional[str] = ".",
    factory: bool = False,
) -> None:
    """
    Starts a Shiny app. Press ``Ctrl+C`` (or ``Ctrl+Break`` on Windows) to stop.

    Parameters
    ----------
    app
        The app to run. The default, ``app:app``, represents the "usual" case where the
        application is named ``app`` inside a ``app.py`` file within the current working
        directory. In other cases, the app location can be specified as a
        ``<module>:<attribute>`` string where the ``:<attribute>`` is only necessary if
        the application is named something other than ``app``. Note that ``<module>``
        can be relative path to a ``.py`` file or a directory (with an ``app.py`` file
        inside it); and in this case, the relative path is resolved relative to the
        ``app_dir`` directory.
    host
        The address that the app should listen on.
    port
        The port that the app should listen on.
    debug
        Enable debug mode.
    reload
        Enable auto-reload.
    ws_max_size
        WebSocket max size message in bytes.
    log_level
        Log level.
    app_dir
        Look for ``app`` under this directory (by adding this to the ``PYTHONPATH``).

    Tip
    ---
    The ``shiny run`` command-line interface (which comes installed with Shiny) provides
    the same functionality as this function.

    Examples
    --------

    .. code-block:: python

        from shiny import run_app

        # Run ``app`` inside ``./app.py``
        run_app()

        # Run ``app`` inside ``./myapp.py`` (or ``./myapp/app.py``)
        run_app("myapp")

        # Run ``my_app`` inside ``./myapp.py`` (or ``./myapp/app.py``)
        run_app("myapp:my_app")

        # Run ``my_app`` inside ``../myapp.py`` (or ``../myapp/app.py``)
        run_app("myapp:my_app", app_dir="..")
    """

    if isinstance(app, str):
        app, app_dir = resolve_app(app, app_dir)

    if app_dir:
        app_dir = os.path.realpath(app_dir)

    log_config: Dict[str, Any] = copy.deepcopy(uvicorn.config.LOGGING_CONFIG)

    if reload and app_dir is not None:
        reload_dirs = [app_dir]
    else:
        reload_dirs = []

    if reload and autoreload_port != "":
        m = re.search("^([+-]?)(\\d+)$", autoreload_port)
        if not m:
            sys.stderr.write(
                "Error: Couldn't understand the provided value for --autoreload-port\n"
            )
            exit(1)
        autoreload_port_num = int(m.group(2))
        if m.group(1) == "+":
            autoreload_port_num += port
        elif m.group(1) == "-":
            autoreload_port_num = port - autoreload_port_num

        if autoreload_port_num == port:
            sys.stderr.write(
                "Autoreload port is already being used by the app; disabling autoreload\n"
            )
        else:
            setup_hot_reload(log_config, autoreload_port_num, port)

    maybe_setup_rsw_proxying(log_config)

    uvicorn.run(
        app,  # type: ignore
        host=host,
        port=port,
        debug=debug,
        reload=reload,
        reload_dirs=reload_dirs,
        ws_max_size=ws_max_size,
        log_level=log_level,
        log_config=log_config,
        app_dir=app_dir,
        factory=factory,
    )


def setup_hot_reload(
    log_config: Dict[str, Any], autoreload_port: int, app_port: int
) -> None:
    # The only way I've found to get notified when uvicorn decides to reload, is by
    # inserting a custom log handler.
    log_config["handlers"]["shiny_hot_reload"] = {
        "class": "shiny._autoreload.HotReloadHandler",
        "level": "INFO",
    }
    log_config["loggers"]["uvicorn.error"]["handlers"] = ["shiny_hot_reload"]

    _autoreload.start_server(autoreload_port, app_port)


def maybe_setup_rsw_proxying(log_config: Dict[str, Any]) -> None:
    # Replace localhost URLs emitted to the log, with proxied URLs
    if _hostenv.is_workbench():
        if "filters" not in log_config:
            log_config["filters"] = {}
        log_config["filters"]["rsw_proxy"] = {"()": "shiny._hostenv.ProxyUrlFilter"}
        if "filters" not in log_config["handlers"]["default"]:
            log_config["handlers"]["default"]["filters"] = []
        log_config["handlers"]["default"]["filters"].append("rsw_proxy")


def is_file(app: str) -> bool:
    return "/" in app or app.endswith(".py")


def resolve_app(app: str, app_dir: Optional[str]) -> Tuple[str, Optional[str]]:
    # The `app` parameter can be:
    #
    # - A module:attribute name
    # - An absolute or relative path to a:
    #   - .py file (look for app inside of it)
    #   - directory (look for app:app inside of it)
    # - A module name (look for :app) inside of it

    module, _, attr = app.partition(":")
    if not module:
        raise ImportError("The APP parameter cannot start with ':'.")
    if not attr:
        attr = "app"

    if is_file(module):
        # TODO: We should probably be using some kind of loader
        # TODO: I don't like that we exit here, if we ever export this it would be bad;
        #       but also printing a massive stack trace for a `shiny run badpath` is way
        #       unfriendly. We should probably throw a custom error that the shiny run
        #       entrypoint knows not to print the stack trace for.
        if not os.path.exists(module):
            sys.stderr.write(f"Error: {module} not found\n")
            sys.exit(1)
        if not os.path.isfile(module):
            sys.stderr.write(f"Error: {module} is not a file\n")
            sys.exit(1)
        dirname, filename = os.path.split(module)
        module = filename[:-3] if filename.endswith(".py") else filename
        app_dir = dirname

    return f"{module}:{attr}", app_dir


def try_import_module(module: str) -> Optional[types.ModuleType]:
    try:
        if not importlib.util.find_spec(module):
            return None
    except ModuleNotFoundError:
        # find_spec throws this when the module contains both '/' and '.' characters
        return None
    except ImportError:
        # find_spec throws this when the module starts with "."
        return None

    # It's important for ModuleNotFoundError and ImportError (and any other error) NOT
    # to be caught here, as we want to report the true error to the user. Otherwise,
    # missing dependencies can be misreported to the user as the app module itself not
    # being found.
    return importlib.import_module(module)


@main.command(
    help="""Create a Shiny application from a template

APPDIR is the directory to the Shiny application. A file named app.py will be created in
that directory.

After creating the application, you use `shiny run`:

    shiny run APPDIR/app.py --reload
"""
)
@click.argument("appdir", type=str)
def create(appdir: str) -> None:
    app_dir = Path(appdir)
    app_path = app_dir / "app.py"
    if app_path.exists():
        print(f"Error: Can't create {app_path} because it already exists.")
        sys.exit(1)

    if not app_dir.exists():
        app_dir.mkdir()

    shutil.copyfile(
        Path(__file__).parent / "examples" / "template" / "app.py", app_path
    )

    print(f"Created Shiny app at {app_dir / 'app.py'}")


@main.command(
    help="""Create a statically deployable distribution for a Shiny app.

APPDIR is the directory containing the Shiny application.

DESTDIR is the destination directory where the output files will be written to. This
directory can be deployed as a static web site.

This will not deploy the contents of venv/ or any files that start with '.'

After writing the output files, you can serve them locally with the following command:

    python3 -m http.server --directory DESTDIR 8008
"""
)
@click.argument("appdir", type=str)
@click.argument("destdir", type=str)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite existing files in the destination directory.",
    show_default=True,
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Print debugging information when copying files.",
    show_default=True,
)
@click.option(
    "--subdir",
    type=str,
    default=None,
    help="Subdir in which to put the app.",
    show_default=True,
)
def static(
    appdir: str, destdir: str, overwrite: bool, subdir: str, verbose: bool
) -> None:
    _static.deploy_static(
        appdir, destdir, overwrite=overwrite, subdir=subdir, verbose=verbose
    )


@main.command(
    help="""Manage local copy of assets for static app deployment.

    \b
    Commands:
        download: Download assets from the remote server.
        remove: Remove local copies of assets.
        info: Print information about the local assets.
        copy: Copy shinylive assets from a local directory. Must be used with --source.

"""
)
@click.argument("command", type=str)
@click.option(
    "--version",
    type=str,
    default=_static._SHINYLIVE_DEFAULT_VERSION,
    help="Shinylive version to download or remove.",
    show_default=True,
)
@click.option(
    "--url",
    type=str,
    default=_static._SHINYLIVE_DOWNLOAD_URL,
    help="URL to download from.",
    show_default=True,
)
@click.option(
    "--dir",
    type=str,
    default=None,
    help="Directory to store shinylive assets (if not using the default)",
)
@click.option(
    "--source",
    type=str,
    default=None,
    help="Directory where shinylive assets will be copied from. Must be used with 'copy' command.",
)
def static_assets(
    command: str, version: str, url: str, dir: str, source: Optional[str]
) -> None:
    if dir is None:
        dir = _static.get_default_shinylive_dir()

    if command == "download":
        print(f"Downloading shinylive-{version} from {url} to {dir}")
        _static.download_shinylive(destdir=dir, version=version, url=url)
    elif command == "remove":
        print(f"Removing shinylive-{version} from {dir}")
        _static.remove_shinylive_local(shinylive_dir=dir, version=version)
    elif command == "info":
        _static.print_shinylive_local_info()
    elif command == "copy":
        if source is None:
            raise click.UsageError("Must specify --source")
        print(f"Copying shinylive-{version} from {source} to {dir}/shinylive-{version}")
        _static.copy_shinylive_local(source_dir=source, destdir=dir, version=version)
    else:
        raise click.UsageError(f"Unknown command: {command}")



import asyncio
from abc import ABC, abstractmethod
from typing import Optional

import starlette.websockets
from starlette.websockets import WebSocketState
from starlette.requests import HTTPConnection


class Connection(ABC):
    """Abstract class to serve a session and send/receive messages to the
    client."""

    @abstractmethod
    async def send(self, message: str) -> None:
        ...

    @abstractmethod
    async def receive(self) -> str:
        ...

    @abstractmethod
    async def close(self, code: int, reason: Optional[str]) -> None:
        ...

    @abstractmethod
    def get_http_conn(self) -> HTTPConnection:
        ...


class MockConnection(Connection):
    def __init__(self):
        # This currently hard-codes some basic values for scope. In the future, we could
        # make those more configurable if we need to customize the HTTPConnection (like
        # "scheme", "path", and "query_string").
        self._http_conn = HTTPConnection(scope={"type": "websocket", "headers": {}})
        self._queue: asyncio.Queue[str] = asyncio.Queue()

    async def send(self, message: str) -> None:
        pass

    async def receive(self) -> str:
        msg = await self._queue.get()
        if msg == "":
            raise ConnectionClosed()
        return msg

    async def close(self, code: int, reason: Optional[str]) -> None:
        pass

    def get_http_conn(self) -> HTTPConnection:
        return self._http_conn

    def cause_receive(self, message: str) -> None:
        """Call from tests to simulate the other side sending a message"""
        self._queue.put_nowait(message)

    def cause_disconnect(self) -> None:
        """Call from tests to simulate the other side disconnecting"""
        self.cause_receive("")


class StarletteConnection(Connection):
    conn: starlette.websockets.WebSocket

    def __init__(self, conn: starlette.websockets.WebSocket):
        self.conn: starlette.websockets.WebSocket = conn
        self._closed = False

    async def accept(self, subprotocol: Optional[str] = None):
        await self.conn.accept(subprotocol)  # type: ignore

    async def send(self, message: str) -> None:
        if self._is_closed():
            return

        try:
            await self.conn.send_text(message)
            return
        # For the record, websockets.exceptions.ConnectionClosed is one exception I see
        # when hammering on the browser reload button
        except Exception:
            # The contract of WebSocket.send() is to never throw (unless the websocket
            # is not yet connected; sending a message after the ws has closed is OK.)
            # However, it's also not very safe to keep using this websocket if we can't
            # be sure they've received this message. So close it, and then continue as
            # if nothing is wrong.
            if not self._is_closed():
                await self.close(1008, "Send failure")
            return

    async def receive(self) -> str:
        if self._is_closed():
            raise ConnectionClosed()

        try:
            return await self.conn.receive_text()
        except starlette.websockets.WebSocketDisconnect:
            raise ConnectionClosed()
        except Exception:
            # From RFC6455:
            # 1008 indicates that an endpoint is terminating the connection because it
            # has received a message that violates its policy.  This is a generic status
            # code that can be returned when there is no other more suitable status code
            # (e.g., 1003 or 1009) or if there is a need to hide specific details about
            # the policy.
            try:
                await self.close(1008, "Receive operation failed")
            except Exception:
                pass
            raise

    async def close(self, code: int, reason: Optional[str]) -> None:
        if self._is_closed():
            return

        # Even if self.conn.close() fails, treat this as closed.
        self._closed = True

        try:
            await self.conn.close(code)
        except Exception:
            # WebSocket failed to close (usually because it already terminated in some
            # unusual fashion); ignoring because the contract of WebSocket.close() is to
            # never throw. (Not in Python, but in the browser, and it's pretty hard to
            # reason about otherwise.)
            pass

    def _is_closed(self) -> bool:
        return (
            self.conn.application_state == WebSocketState.DISCONNECTED  # type: ignore
            or self.conn.client_state == WebSocketState.DISCONNECTED  # type: ignore
            or self._closed
        )

    def get_http_conn(self) -> HTTPConnection:
        return self.conn


class ConnectionClosed(Exception):
    """Raised when a Connection is closed from the other side."""

    pass



"""
Information about the execution environment for Shiny
"""

import sys

# True if we're executing in WASM mode, False if we're executing using actual Python
is_pyodide: bool = "pyodide" in sys.modules



__all__ = (
    "input_slider",
    "SliderValueArg",
    "SliderStepArg",
    "AnimationOptions",
)

import math
import sys
from datetime import date, datetime, timedelta
from typing import Dict, Iterable, Optional, TypeVar, Union, cast

from htmltools import HTML, Tag, TagAttrArg, TagChildArg, css, div, tags

from .._docstring import add_example
from ._html_dependencies import ionrangeslider_deps
from .._namespaces import resolve_id
from ._utils import shiny_input_label

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import NotRequired, TypedDict
else:
    from typing_extensions import NotRequired, TypedDict

# TODO: validate value(s) are within (min,max)?

SliderValueArg = TypeVar("SliderValueArg", float, datetime, date)
SliderStepArg = Union[float, timedelta]


class AnimationOptions(TypedDict):
    """
    Options for the animation of a :func:`input_slider`.

    Parameters
    ----------
    interval
        The interval, in milliseconds, between each animation step.
    loop
        ``True`` to automatically restart the animation when it reaches the end.
    play_button
        Play button text or HTML.
    pause_button
        Pause button text or HTML.

    Returns
    -------
    A TypedDict.

    See Also
    --------
    ~shiny.ui.input_slider
    """

    interval: NotRequired[int]
    loop: NotRequired[bool]
    play_button: NotRequired[TagChildArg]
    pause_button: NotRequired[TagChildArg]


@add_example()
def input_slider(
    id: str,
    label: TagChildArg,
    min: SliderValueArg,
    max: SliderValueArg,
    value: Union[SliderValueArg, Iterable[SliderValueArg]],
    *,
    step: Optional[SliderStepArg] = None,
    ticks: bool = True,
    animate: Union[bool, AnimationOptions] = False,
    width: Optional[str] = None,
    sep: str = ",",
    pre: Optional[str] = None,
    post: Optional[str] = None,
    time_format: Optional[str] = None,
    timezone: Optional[str] = None,
    drag_range: bool = True,
) -> Tag:
    """
    Constructs a slider widget to select a number, date, or date-time from a range.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    value
        Initial value.
    step
        Interval to use when stepping between min and max.
    ticks
        ``False`` to hide tick marks, ``True`` to show them according to some simple
        heuristics.
    animate
        ``True`` to show simple animation controls with default settings; ``False`` not
        to; or a custom settings list, such as those created using
        :class:`AnimationOptions()`.
    width
        The CSS width, e.g. '400px', or '100%'
    sep
        Separator between thousands places in numbers.
    pre
        A prefix string to put in front of the value.
    post
        A suffix string to put after the value.
    time_format
        Only used if the slider values are :func:`~datetime.date` or
        :func:`~datetime.datetime` objects. A time format string, to be passed to the
        Javascript strftime library. See https://github.com/samsonjs/strftime for more
        details. For Dates, the default is "%F" (like "2015-07-01"), and for Datetimes,
        the default is "%F %T" (like "2015-07-01 15:32:10").
    timezone
        Only used if the values are :func:`~datetime.datetime` objects. A string
        specifying the time zone offset for the displayed times, in the format "+HHMM"
        or "-HHMM". If ``None`` (the default), times will be displayed in the browser's
        time zone. The value "+0000" will result in UTC time.
    drag_range
        This option is used only if it is a range slider (with two values). If ``True``
        (the default), the range can be dragged. In other words, the min and max can be
        dragged together. If ``False``, the range cannot be dragged.

    Returns
    -------
    A UI element

    Notes
    ------
    .. admonition:: Server value

       A number, date, or date-time (depending on the class of value), or in the case of
       slider range, a list of two numbers/dates/date-times.

    See Also
    -------
    ~shiny.ui.update_slider
    """

    # Thanks to generic typing, max, value, etc. should be of the same type
    data_type = _slider_type(min)

    # Make sure min, max, value, and step are all numeric
    # (converts dates/datetimes to milliseconds since epoch...this is the value JS wants)
    min_num = _as_numeric(min)
    max_num = _as_numeric(max)
    val_nums = (
        (_as_numeric(value[0]), _as_numeric(value[1]))
        if isinstance(value, (tuple, list))
        else (
            _as_numeric(cast(SliderValueArg, value)),
            _as_numeric(cast(SliderValueArg, value)),
        )
    )
    step_num = _find_step_size(min_num, max_num) if step is None else _as_numeric(step)

    n_ticks = None
    if ticks:
        n_steps = (max_num - min_num) / step_num
        # Make sure there are <= 10 steps.
        # n_ticks can be a noninteger, which is good when the range is not an
        # integer multiple of the step size, e.g., min=1, max=10, step=4
        scale_factor = math.ceil(n_steps / 10)
        n_ticks = n_steps / scale_factor

    id = resolve_id(id)

    props: Dict[str, TagAttrArg] = {
        "class_": "js-range-slider",
        "id": id,
        "style": css(width=width),
        "data_skin": "shiny",
        # TODO: do we need to worry about scientific notation (i.e., formatNoSci()?)
        "data_min": str(min_num),
        "data_max": str(max_num),
        "data_from": str(val_nums[0]),
        "data_step": str(step_num),
        "data_grid": ticks,
        "data_grid_num": n_ticks,
        "data_grid_snap": "false",
        "data_prettify_separator": sep,
        "data_prettify_enabled": sep != "",
        "data_prefix": pre,
        "data_postfix": post,
        "data_keyboard": "true",
        "data_data_type": data_type,
        "data_time_format": time_format,
        "data_timezone": timezone,
    }

    if isinstance(value, (tuple, list)):
        props["data_type"] = "double"
        props["data_to"] = str(val_nums[1])
        props["data_drag_interval"] = drag_range

    if not time_format and data_type[0:4] == "date":
        props["data_time_format"] = "%F" if data_type == "date" else "%F %T"

    # ionRangeSlider wants attr = 'true'/'false'
    props = {k: str(v).lower() if isinstance(v, bool) else v for k, v in props.items()}

    slider_tag = div(
        shiny_input_label(id, label),
        tags.input(**props),
        *ionrangeslider_deps(),
        class_="form-group shiny-input-container",
    )

    if animate is False:
        return slider_tag

    if animate is True:
        animate = AnimationOptions()

    animate_tag = div(
        tags.a(
            tags.span(animate.get("play_button", _play_icon()), class_="play"),
            tags.span(animate.get("pause_button", _pause_icon()), class_="pause"),
            href="#",
            class_="slider-animate-button link-secondary",
            data_target_id=id,
            data_interval=animate.get("interval", 500),
            data_loop=animate.get("loop", True),
        ),
        class_="slider-animate-container",
    )

    slider_tag.append(animate_tag)

    return slider_tag


def _slider_type(x: SliderValueArg) -> str:
    if isinstance(x, datetime):
        return "datetime"
    if isinstance(x, date):
        return "date"
    return "number"


def _as_numeric(x: Union[SliderStepArg, datetime, date]) -> float:
    if isinstance(x, timedelta):
        return x.total_seconds() * 1000
    if isinstance(x, datetime):
        return x.timestamp() * 1000
    if isinstance(x, date):
        return datetime(x.year, x.month, x.day).timestamp() * 1000
    return x


def _find_step_size(
    min: Union[int, float], max: Union[int, float]
) -> Union[int, float]:
    # TODO: this is a naive version of shiny::findStepSize() that might be susceptible to
    # rounding errors? https://github.com/rstudio/shiny/pull/1956
    range = max - min

    if range < 2 or isinstance(min, float) or isinstance(max, float):
        step = range / 100
        # Round the step to get rid of any floating point arithmetic errors by
        # mimicing what signif(digits = 10, step) does in R (see Description of ?signif)
        # (the basic intuition is that smaller differences need more precision)
        return round(step, 10 - math.ceil(math.log10(step)))
    else:
        return 1


play_icon = """<svg viewBox="0 0 448 512" preserveAspectRatio="none" aria-hidden="true" role="img" style="fill:currentColor;height:1em;width:0.88em;margin-left:auto;margin-right:0.2em;position:relative;vertical-align:-0.125em;font-size:inherit;overflow:visible;">
  <title>Play</title>
  <!-- Font Awesome Free 5.15.4 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) -->
  <path d="M424.4 214.7L72.4 6.6C43.8-10.3 0 6.1 0 47.9V464c0 37.5 40.7 60.1 72.4 41.3l352-208c31.4-18.5 31.5-64.1 0-82.6z"></path>
</svg>"""

pause_icon = """<svg viewBox="0 0 448 512" preserveAspectRatio="none" aria-hidden="true" role="img" style="fill:currentColor;height:1em;width:0.88em;margin-left:auto;margin-right:0.2em;position:relative;vertical-align:-0.125em;font-size:inherit;overflow:visible;">
  <title>Pause</title>
  <!-- Font Awesome Free 5.15.4 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) -->
  <path d="M144 479H48c-26.5 0-48-21.5-48-48V79c0-26.5 21.5-48 48-48h96c26.5 0 48 21.5 48 48v352c0 26.5-21.5 48-48 48zm304-48V79c0-26.5-21.5-48-48-48h-96c-26.5 0-48 21.5-48 48v352c0 26.5 21.5 48 48 48h96c26.5 0 48-21.5 48-48z"></path>
</svg>"""


def _play_icon() -> Union[Tag, HTML]:
    return HTML(play_icon)


def _pause_icon() -> Union[Tag, HTML]:
    return HTML(pause_icon)



"""Tests for """


import random
import textwrap
from typing import Callable, Any

from shiny import ui
from shiny.ui._navs import NavSet
from shiny._utils import private_seed
from htmltools import TagList


# Fix the randomness of these functions to make the tests deterministic
def with_private_seed(func: Callable[[], NavSet], *args: Any, **kwargs: Any):
    with private_seed():
        random.seed(0)
        return func(*args, **kwargs)


def test_nav_markup():
    a = ui.nav("a", "a")
    b = ui.nav("b", "b")
    c = ui.nav("c", "c")
    menu = ui.nav_menu(
        "Menu",
        c,
        "----",
        "Plain text",
        "----",
        ui.nav_control("Other item"),
    )

    x = with_private_seed(ui.navset_tab, a, b, ui.nav_control("Some item"), menu)

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <ul class="nav nav-tabs" data-tabsetid="7311">
          <li class="nav-item">
            <a data-bs-toggle="tab" data-toggle="tab" data-value="a" role="tab" class="nav-link active" href="#tab-7311-0">a</a>
          </li>
          <li class="nav-item">
            <a data-bs-toggle="tab" data-toggle="tab" data-value="b" role="tab" class="nav-link" href="#tab-7311-1">b</a>
          </li>
          <li>Some item</li>
          <li class="nav-item dropdown">
            <a class="nav-link dropdown-toggle " data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
            <ul class="dropdown-menu " data-tabsetid="7890">
              <li>
                <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item" href="#tab-7890-0">c</a>
              </li>
              <li class="dropdown-divider"></li>
              <li class="dropdown-header">Plain text</li>
              <li class="dropdown-divider"></li>
              <li>Other item</li>
            </ul>
          </li>
        </ul>
        <div class="tab-content" data-tabsetid="7311">
          <div class="tab-pane active" role="tabpanel" data-value="a" id="tab-7311-0">a</div>
          <div class="tab-pane" role="tabpanel" data-value="b" id="tab-7311-1">b</div>
          <div class="tab-pane" role="tabpanel" data-value="c" id="tab-7890-0">c</div>
        </div>"""
    )

    x = with_private_seed(
        ui.navset_pill,
        menu,
        a,
        id="navset_pill_id",
    )

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <ul class="nav nav-pills shiny-tab-input" id="navset_pill_id" data-tabsetid="7311">
          <li class="nav-item dropdown">
            <a class="nav-link dropdown-toggle active" data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
            <ul class="dropdown-menu " data-tabsetid="7890">
              <li>
                <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item active" href="#tab-7890-0">c</a>
              </li>
              <li class="dropdown-divider"></li>
              <li class="dropdown-header">Plain text</li>
              <li class="dropdown-divider"></li>
              <li>Other item</li>
            </ul>
          </li>
          <li class="nav-item">
            <a data-bs-toggle="tab" data-toggle="tab" data-value="a" role="tab" class="nav-link" href="#tab-7311-1">a</a>
          </li>
        </ul>
        <div class="tab-content" data-tabsetid="7311">
          <div class="tab-pane active" role="tabpanel" data-value="c" id="tab-7890-0">c</div>
          <div class="tab-pane" role="tabpanel" data-value="a" id="tab-7311-1">a</div>
        </div>"""
    )

    x = with_private_seed(
        ui.navset_pill_card,
        a,
        ui.nav_menu("Menu", c),
        b,
        selected="c",
    )

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <div class="card">
          <div class="card-header">
            <ul class="nav nav-pills card-header-pills" data-tabsetid="7311">
              <li class="nav-item">
                <a data-bs-toggle="tab" data-toggle="tab" data-value="a" role="tab" class="nav-link" href="#tab-7311-0">a</a>
              </li>
              <li class="nav-item dropdown">
                <a class="nav-link dropdown-toggle active" data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
                <ul class="dropdown-menu " data-tabsetid="7890">
                  <li>
                    <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item active" href="#tab-7890-0">c</a>
                  </li>
                </ul>
              </li>
              <li class="nav-item">
                <a data-bs-toggle="tab" data-toggle="tab" data-value="b" role="tab" class="nav-link" href="#tab-7311-2">b</a>
              </li>
            </ul>
          </div>
          <div class="card-body">
            <div class="tab-content" data-tabsetid="7311">
              <div class="tab-pane" role="tabpanel" data-value="a" id="tab-7311-0">a</div>
              <div class="tab-pane active" role="tabpanel" data-value="c" id="tab-7890-0">c</div>
              <div class="tab-pane" role="tabpanel" data-value="b" id="tab-7311-2">b</div>
            </div>
          </div>
        </div>"""
    )

    x = with_private_seed(
        ui.navset_bar,  # type: ignore
        ui.nav_menu("Menu", "Plain text", c),
        title="Page title",
        footer="Page footer",
        header="Page header",
    )

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <nav class="navbar navbar-expand-md navbar-light bg-light">
          <div class="container-fluid">
            <a class="navbar-brand" href="#">Page title</a>
            <button class="navbar-toggler" type="button" data-bs-toggle="collapse" data-bs-target="#navbar-collapse-1663" aria-controls="navbar-collapse-1663" aria-expanded="false" aria-label="Toggle navigation">
              <span class="navbar-toggler-icon"></span>
            </button>
            <div id="navbar-collapse-1663" class="collapse navbar-collapse">
              <ul class="nav navbar-nav" data-tabsetid="7311">
                <li class="nav-item dropdown">
                  <a class="nav-link dropdown-toggle active" data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
                  <ul class="dropdown-menu " data-tabsetid="7890">
                    <li class="dropdown-header">Plain text</li>
                    <li>
                      <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item active" href="#tab-7890-1">c</a>
                    </li>
                  </ul>
                </li>
              </ul>
            </div>
          </div>
        </nav>
        <div class="container-fluid">
          <div class="row">Page header</div>
          <div class="tab-content" data-tabsetid="7311">
            <div class="tab-pane active" role="tabpanel" data-value="c" id="tab-7890-1">c</div>
          </div>
          <div class="row">Page footer</div>
        </div>"""
    )



"""Tests for polling-related functionality."""

import os
import tempfile
from enum import Enum
from random import random
from types import TracebackType
from typing import Any, Dict, Callable, Optional, Type, cast

import pytest

from shiny import *
from shiny import _utils
from shiny.reactive import *
from shiny._namespaces import Root

from .mocktime import MockTime


class OnEndedSessionCallbacks:
    """
    A far-too-minimal mock of Session that implements nothing but on_ended. This is
    used so that invalidate_later calls can be cleaned up, otherwise you get warnings
    about pending tasks when pytest completes.

    Eventually we should have a proper mock of Session, then we can retire this.
    """

    ns = Root

    def __init__(self):
        self._on_ended_callbacks = _utils.Callbacks()
        # Unfortunately we have to lie here and say we're a session. Obvously, any
        # attempt to call anything but session.on_ended() will fail.
        self._session_context = session.session_context(cast(Session, self))

    def on_ended(self, fn: Callable[[], None]) -> Callable[[], None]:
        return self._on_ended_callbacks.register(fn)

    def _send_message_sync(self, message: Dict[str, object]) -> None:
        pass

    async def __aenter__(self):
        self._session_context.__enter__()

    async def __aexit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ):
        self._session_context.__exit__(exctype, excinst, exctb)
        self._on_ended_callbacks.invoke()


@pytest.mark.asyncio
async def test_poll():
    async with OnEndedSessionCallbacks():
        poll_invocations = 0
        poll_return1 = 0  # A non-reactive component of the return value
        poll_return2 = Value(0)  # A reactive component of the return value
        value_invocations = 0
        value_dep = Value(0)

        def poll_func():
            nonlocal poll_invocations
            poll_invocations += 1
            return poll_return1 + poll_return2()

        @poll(poll_func)
        def value_func():
            value_dep()  # Take a reactive dependency on value_dep
            nonlocal value_invocations
            value_invocations += 1

        mock_time = MockTime()
        with mock_time():
            # Poll func is invoked once during @poll(), to seed the underlying reactive val
            assert (poll_invocations, value_invocations) == (1, 0)

            await flush()

            # The observer that updates poll has executed once.
            # @poll returns a lazy Calc, so value hasn't been invoked yet.
            assert (poll_invocations, value_invocations) == (2, 0)

            @Effect()
            def _():
                value_func()

            await flush()
            assert (poll_invocations, value_invocations) == (2, 1)

            await flush()
            assert (poll_invocations, value_invocations) == (2, 1)

            await mock_time.advance_time(1.01)
            # poll_invocations advances without a flush() because invalidate_later itself
            # invokes flush()
            assert (poll_invocations, value_invocations) == (3, 1)
            await flush()
            assert (poll_invocations, value_invocations) == (3, 1)

            # Now change value
            poll_return1 += 1
            await flush()
            # Nothing changes because no time has passed, poll() has not tried again
            assert (poll_invocations, value_invocations) == (3, 1)
            await mock_time.advance_time(1.01)
            assert (poll_invocations, value_invocations) == (4, 2)

            # When a reactive dependency of poll_func invalidates, there's no need to wait
            # until the next poll
            with isolate():
                poll_return2.set(poll_return2() + 1)
            await flush()
            assert (poll_invocations, value_invocations) == (5, 3)

            # When a reactive dependency of value_func invalidates, there's no need to wait
            # until the next poll
            with isolate():
                value_dep.set(value_dep() + 1)
            await flush()
            assert (poll_invocations, value_invocations) == (5, 4)

            await mock_time.advance_time(1.01)
            assert (poll_invocations, value_invocations) == (6, 4)


@pytest.mark.asyncio
async def test_poll_errors():
    async with OnEndedSessionCallbacks():

        class Mode(Enum):
            NORMAL = 0
            RAISE = 1
            INCOMPARABLE = 2

        class Incomparable:
            def __eq__(self, other: Any):
                raise ValueError("I refused to be compared to anyone!")

        mock_time = MockTime()
        with mock_time():
            mode = Mode.NORMAL

            def poll_func() -> Any:
                if mode is Mode.NORMAL:
                    return random()
                if mode is Mode.RAISE:
                    raise ValueError("boom")
                if mode is Mode.INCOMPARABLE:
                    return Incomparable()

            @poll(poll_func)
            def bad_poll() -> Any:
                return random()

            invocations = 0

            @Effect()
            def _():
                nonlocal invocations
                invocations += 1
                if mode is Mode.NORMAL:
                    bad_poll()
                if mode is Mode.RAISE:
                    with pytest.raises(ValueError):
                        bad_poll()
                if mode is Mode.INCOMPARABLE:
                    with pytest.raises(TypeError):
                        bad_poll()

            with isolate():
                await flush()
                bad_poll()

                mode = Mode.RAISE
                # Doesn't raise because polling hasn't happened yet
                bad_poll()

                await mock_time.advance_time(1.01)
                await flush()

                with pytest.raises(ValueError):
                    bad_poll()

                mode = Mode.INCOMPARABLE
                await mock_time.advance_time(1.01)
                await flush()

                with pytest.raises(TypeError):
                    bad_poll()

                assert invocations == 3


@pytest.mark.asyncio
async def test_file_reader():
    tmpfile = tempfile.NamedTemporaryFile(delete=False)
    try:
        async with OnEndedSessionCallbacks():
            invocations = 0

            tmpfile.write("hello\n".encode())
            tmpfile.flush()
            tmpfile.close()

            mock_time = MockTime()
            with mock_time():

                @file_reader(tmpfile.name)
                def read_file():
                    with open(tmpfile.name, "r") as f:
                        nonlocal invocations
                        invocations += 1
                        return f.read()

                with isolate():
                    assert invocations == 0
                    await flush()

                    assert read_file() == "hello\n"
                    assert invocations == 1
                    # Advancing time without a write does nothing
                    await mock_time.advance_time(1.01)
                    assert read_file() == "hello\n"
                    assert invocations == 1

                    with open(tmpfile.name, "a") as f:
                        f.write("goodbye\n")

                    # The file's been updated, but we haven't looked yet
                    assert read_file() == "hello\n"

                    await mock_time.advance_time(1.01)

                    assert read_file() == "hello\ngoodbye\n"
                    assert invocations == 2
    finally:
        os.unlink(tmpfile.name)


@pytest.mark.asyncio
async def test_file_reader_error():
    async with OnEndedSessionCallbacks():

        tmpfile1 = tempfile.NamedTemporaryFile(delete=False)
        mock_time = MockTime()
        with mock_time():

            @file_reader(tmpfile1.name)
            def read_file():
                return True

            with isolate():
                await flush()
                assert read_file() is True
                tmpfile1.close()

                os.unlink(tmpfile1.name)
                await mock_time.advance_time(1.01)
                await flush()

                with pytest.raises(FileNotFoundError):
                    read_file()



__all__ = (
    "input_checkbox",
    "input_checkbox_group",
    "input_radio_buttons",
)

from typing import List, Mapping, Optional, Union

from htmltools import Tag, TagChildArg, css, div, span, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._utils import shiny_input_label

# Canonical format for representing select options.
_Choices = Mapping[str, TagChildArg]

# Formats available to the user
ChoicesArg = Union[List[str], _Choices]


@add_example()
def input_checkbox(
    id: str, label: TagChildArg, value: bool = False, *, width: Optional[str] = None
) -> Tag:
    """
    Create a checkbox that can be used to specify logical values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    width
        The CSS width, e.g. '400px', or '100%'

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        ``True`` if checked, ``False`` otherwise.

    See Also
    -------
    ~shiny.ui.input_switch
    ~shiny.ui.update_checkbox
    ~shiny.ui.input_checkbox_group
    ~shiny.ui.input_radio_buttons
    """

    return div(
        div(
            tags.label(
                tags.input(
                    id=resolve_id(id),
                    type="checkbox",
                    checked="checked" if value else None,
                ),
                span(label),
            ),
            class_="checkbox",
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )


def input_switch(
    id: str, label: TagChildArg, value: bool = False, *, width: Optional[str] = None
) -> Tag:
    """
    Create a switch that can be used to specify logical values. Similar to
    ~shiny.ui.input_checkbox, but implies to the user that the change will take effect
    immediately.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    width
        The CSS width, e.g. '400px', or '100%'

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        ``True`` if checked, ``False`` otherwise.

    See Also
    -------
    ~shiny.ui.input_checkbox
    ~shiny.ui.update_switch
    ~shiny.ui.input_checkbox_group
    ~shiny.ui.input_radio_buttons
    """

    return _input_checkbox(id, label, "form-check form-switch", value, width=width)


def _input_checkbox(
    id: str,
    label: TagChildArg,
    class_: str = "form-check",
    value: bool = False,
    *,
    width: Optional[str] = None,
) -> Tag:
    return div(
        div(
            tags.input(
                id=resolve_id(id),
                class_="form-check-input",
                type="checkbox",
                checked="checked" if value else None,
            ),
            tags.label(label, class_="form-check-label", for_=resolve_id(id)),
            class_=class_,
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )


@add_example()
def input_checkbox_group(
    id: str,
    label: TagChildArg,
    choices: ChoicesArg,
    *,
    selected: Optional[Union[str, List[str]]] = None,
    inline: bool = False,
    width: Optional[str] = None,
) -> Tag:
    """
    Create a group of checkboxes that can be used to toggle multiple choices
    independently.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels.
    selected
        The values that should be initially selected, if any.
    inline
        If `True`, the result is displayed inline
    width
        The CSS width, e.g. '400px', or '100%'

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        A tuple of string(s) with the selected value(s) (if any).

    See Also
    -------
    ~shiny.ui.update_checkbox_group
    ~shiny.ui.input_checkbox
    ~shiny.ui.input_radio_buttons
    """

    input_label = shiny_input_label(id, label)
    options = _generate_options(
        id=resolve_id(id),
        type="checkbox",
        choices=choices,
        selected=selected,
        inline=inline,
    )
    return div(
        input_label,
        options,
        id=resolve_id(id),
        style=css(width=width),
        class_="form-group shiny-input-checkboxgroup shiny-input-container"
        + (" shiny-input-container-inline" if inline else ""),
        # https://www.w3.org/TR/wai-aria-practices/examples/checkbox/checkbox-1/checkbox-1.html
        role="group",
        aria_labelledby=input_label.attrs.get("id"),
    )


@add_example()
def input_radio_buttons(
    id: str,
    label: TagChildArg,
    choices: ChoicesArg,
    *,
    selected: Optional[str] = None,
    inline: bool = False,
    width: Optional[str] = None,
) -> Tag:
    """
    Create a set of radio buttons used to select an item from a list.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels.
    selected
        The values that should be initially selected, if any.
    inline
        If ``True``, the result is displayed inline
    width
        The CSS width, e.g. '400px', or '100%'
    Returns
    -------
    A UI element

    Notes
    ------
    .. admonition:: Server value

        A string with the selected value.

    See Also
    -------
    ~shiny.ui.update_radio_buttons
    ~shiny.ui.input_checkbox_group
    ~shiny.ui.input_checkbox
    """

    input_label = shiny_input_label(id, label)
    options = _generate_options(
        id=resolve_id(id),
        type="radio",
        choices=choices,
        selected=selected,
        inline=inline,
    )
    return div(
        input_label,
        options,
        id=resolve_id(id),
        style=css(width=width),
        class_="form-group shiny-input-radiogroup shiny-input-container"
        + (" shiny-input-container-inline" if inline else ""),
        # https://www.w3.org/TR/2017/WD-wai-aria-practices-1.1-20170628/examples/radio/radio-1/radio-1.html
        role="radiogroup",
        aria_labelledby=input_label.attrs.get("id"),
    )


def _generate_options(
    id: str,
    type: str,
    choices: ChoicesArg,
    selected: Optional[Union[str, List[str]]],
    inline: bool,
) -> Tag:
    choicez = _normalize_choices(choices)

    if selected is None:
        if type == "radio":
            selected = list(choicez.keys())[0]
        else:
            selected = []

    if not isinstance(selected, list):
        selected = [selected]

    return div(
        [
            _generate_option(
                id,
                type,
                value=choice[0],
                label=choice[1],
                checked=choice[0] in selected,
                inline=inline,
            )
            for choice in choicez.items()
        ],
        class_="shiny-options-group",
    )


def _generate_option(
    id: str,
    type: str,
    value: str,
    label: TagChildArg,
    checked: bool,
    inline: bool,
) -> Tag:
    input = tags.input(
        type=type,
        name=id,
        value=value,
        checked="checked" if checked else None,
    )
    if inline:
        return tags.label(input, span(label), class_=type + "-inline")
    else:
        return div(tags.label(input, span(label)), class_=type)


def _normalize_choices(x: ChoicesArg) -> _Choices:
    if isinstance(x, list):
        return {k: k for k in x}
    else:
        return x



import asyncio
import html
import http
import logging
import os
import secrets
from typing import Optional, Tuple
import threading

from asgiref.typing import (
    ASGI3Application,
    Scope,
    ASGIReceiveCallable,
    ASGISendCallable,
    ASGISendEvent,
)

from ._shinyenv import is_pyodide

if not is_pyodide:
    # Warning: `import websockets` alone doesn't work because the websockets package uses
    # a lazy loading technique which is opaque to mypy and pyright.
    # https://github.com/aaugustin/websockets/issues/940#issuecomment-874012438
    from websockets.client import connect
    from websockets.server import serve, WebSocketServerProtocol
    import websockets.exceptions
    import websockets.datastructures

from ._hostenv import get_proxy_url

# CHILD PROCESS ------------------------------------------------------------


def autoreload_url() -> Optional[str]:
    port = os.getenv("SHINY_AUTORELOAD_PORT")
    if not port:
        return None
    else:
        return get_proxy_url(f"ws://127.0.0.1:{port}/autoreload")


# Instantiated by Uvicorn in worker process when reload is enabled
class HotReloadHandler(logging.Handler):
    """Uvicorn log reader, detects reload events."""

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record: logging.LogRecord) -> None:
        # https://github.com/encode/uvicorn/blob/266db48888f3f1ba56710a49ec82e12eecde3aa3/uvicorn/supervisors/statreload.py#L46
        # https://github.com/encode/uvicorn/blob/266db48888f3f1ba56710a49ec82e12eecde3aa3/uvicorn/supervisors/watchgodreload.py#L148
        if "Reloading..." in record.getMessage():
            reload_begin()
        # https://github.com/encode/uvicorn/blob/926a8f5dc2c9265d5fb5daaafce13878f477e264/uvicorn/lifespan/on.py#L59
        elif "Application startup complete." in record.getMessage():
            reload_end()


# Called from child process when old application instance is shut down
def reload_begin():
    pass


# Called from child process when new application instance starts up
def reload_end():
    # os.kill(os.getppid(), signal.SIGUSR1)

    port = os.getenv("SHINY_AUTORELOAD_PORT")
    if not port:
        return None

    url = f"ws://127.0.0.1:{port}/notify"

    async def _() -> None:
        options = {
            "extra_headers": {
                "Shiny-Autoreload-Secret": os.getenv("SHINY_AUTORELOAD_SECRET", ""),
            }
        }
        try:
            async with connect(url, **options) as websocket:
                await websocket.send("reload_end")
        except websockets.exceptions.ConnectionClosed:
            pass

    asyncio.create_task(_())


class InjectAutoreloadMiddleware:
    """Inserts shiny-autoreload.js into the head.

    It's necessary to do it using middleware instead of in a nice htmldependency,
    because we want autoreload to be effective even when displaying an error page.
    """

    def __init__(self, app: ASGI3Application):
        self.app = app
        ws_url = autoreload_url()
        self.script = (
            f"""  <script src="__shared/shiny-autoreload.js" data-ws-url="{html.escape(ws_url)}"></script>
</head>""".encode(
                "ascii"
            )
            if ws_url
            else bytes()
        )

    async def __call__(
        self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
    ) -> None:
        if scope["type"] != "http" or scope["path"] != "/" or len(self.script) == 0:
            return await self.app(scope, receive, send)
        intercept = True
        body = b""

        async def rewrite_send(event: ASGISendEvent) -> None:
            nonlocal intercept
            nonlocal body

            if intercept:
                if event["type"] == "http.response.start":
                    # Must remove Content-Length, if present; if we insert our
                    # scripts, it won't be correct anymore
                    event["headers"] = [
                        (name, value)
                        for (name, value) in event["headers"]
                        if name.decode("ascii").lower() != "content-length"
                    ]
                elif event["type"] == "http.response.body":
                    body += event["body"]
                    if b"</head>" in body:
                        event["body"] = body.replace(b"</head>", self.script, 1)
                        body = b""  # Allow gc
                        intercept = False
                    elif "more_body" in event and event["more_body"]:
                        # DO NOT send the response; wait for more data
                        return
                    else:
                        # The entire response was seen, and we never encountered
                        # any </head>. Just send everything we have
                        event["body"] = body
                        body = b""  # Allow gc

            return await send(event)

        await self.app(scope, receive, rewrite_send)


# PARENT PROCESS ------------------------------------------------------------


def start_server(port: int, app_port: int):
    """Starts a websocket server that listens on its own port (separate from the main
    Shiny listener).

    Clients can connect on either the /autoreload or /notify path.

    Clients from the uvicorn worker process connect to the /notify path to notify us
    that a successful startup or reload has occured.

    Clients from browsers (on localhost only) connect to the /autoreload path to be
    notified when a successful startup or reload has occured.
    """

    # Store port and secret in environment variables so they are inherited by uvicorn
    # worker processes.
    secret = secrets.token_hex(32)
    os.environ["SHINY_AUTORELOAD_PORT"] = str(port)
    os.environ["SHINY_AUTORELOAD_SECRET"] = secret

    app_url = get_proxy_url(f"http://127.0.0.1:{app_port}/")

    # Run on a background thread so our event loop doesn't interfere with uvicorn.
    # Set daemon=True because we don't want to keep the process alive with this thread.
    threading.Thread(
        None, _thread_main, args=[port, app_url, secret], daemon=True
    ).start()


def _thread_main(port: int, app_url: str, secret: str):
    asyncio.run(_coro_main(port, app_url, secret))


async def _coro_main(port: int, app_url: str, secret: str) -> None:
    reload_now: asyncio.Event = asyncio.Event()

    def nudge():
        reload_now.set()
        reload_now.clear()

    async def reload_server(conn: WebSocketServerProtocol):
        try:
            if conn.path == "/autoreload":
                # The client wants to be notified when the app has reloaded. The client
                # in this case is the web browser, specifically shiny-autoreload.js.
                while True:
                    await reload_now.wait()
                    await conn.send("autoreload")
            elif conn.path == "/notify":
                # The client is notifying us that the app has reloaded. The client in
                # this case is the uvicorn worker process (see reload_end(), above).
                req_secret = conn.request_headers.get("Shiny-Autoreload-Secret", "")
                if req_secret != secret:
                    # The client coudn't prove that they were from a child process
                    return
                data = await conn.recv()
                if isinstance(data, str) and data == "reload_end":
                    nudge()
        except websockets.exceptions.ConnectionClosed:
            pass

    # Handle non-WebSocket requests to the autoreload HTTP server. Without this, if you
    # happen to visit the autoreload endpoint in a browser, you get an error message
    # about only WebSockets being supported. This is not an academic problem as the
    # VSCode extension used in RSW sniffs out ports that are being listened on, which
    # leads to confusion if all you get is an error.
    async def process_request(
        path: str, request_headers: websockets.datastructures.Headers
    ) -> Optional[Tuple[http.HTTPStatus, websockets.datastructures.HeadersLike, bytes]]:
        # If there's no Upgrade header, it's not a WebSocket request.
        if request_headers.get("Upgrade") is None:
            return (http.HTTPStatus.MOVED_PERMANENTLY, [("Location", app_url)], b"")

    async with serve(reload_server, "127.0.0.1", port, process_request=process_request):
        await asyncio.Future()  # wait forever



__all__ = ("input_password",)

from typing import Optional

from htmltools import tags, Tag, div, css, TagChildArg

from .._docstring import add_example
from .._namespaces import resolve_id
from ._utils import shiny_input_label


@add_example()
def input_password(
    id: str,
    label: TagChildArg,
    value: str = "",
    *,
    width: Optional[str] = None,
    placeholder: Optional[str] = None,
) -> Tag:
    """
    Create an password control for entry of passwords.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    width
        The CSS width, e.g. '400px', or '100%'
    placeholder
        The placeholder of the input.

    Returns
    -------
    A UI element.

    Notes
    ------
    .. admonition:: Server value

        A string of the password input. The default value is unless value is provided.

    See Also
    -------
    ~shiny.ui.update_text
    """
    return div(
        shiny_input_label(id, label),
        tags.input(
            id=resolve_id(id),
            type="password",
            value=value,
            class_="form-control",
            placeholder=placeholder,
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )



from shiny import *

app_ui = ui.page_fluid(
    ui.layout_sidebar(
        ui.panel_sidebar(
            ui.tags.p("The first slider controls the second"),
            ui.input_slider("control", "Controller:", min=0, max=20, value=10, step=1),
            ui.input_slider("receive", "Receiver:", min=0, max=20, value=10, step=1),
        ),
        ui.panel_main(),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        val = input.control()
        # Control the value, min, max, and step.
        # Step size is 2 when input value is even; 1 when value is odd.
        ui.update_slider(
            "receive", value=val, min=int(val / 2), max=val + 4, step=(val + 1) % 2 + 1
        )


app = App(app_ui, server)



import asyncio
import contextlib
import functools
import importlib
import inspect
import mimetypes
import os
import random
import secrets
import sys
import tempfile
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard

if sys.version_info >= (3, 8):
    CancelledError = asyncio.CancelledError
else:
    CancelledError = asyncio.futures.CancelledError

# ==============================================================================
# Misc utility functions
# ==============================================================================
def rand_hex(bytes: int) -> str:
    """
    Creates a random hexadecimal string of size `bytes`. The length in
    characters will be bytes*2.
    """
    format_str = "{{:0{}x}}".format(bytes * 2)
    return format_str.format(secrets.randbits(bytes * 8))


def drop_none(x: Dict[str, Any]) -> Dict[str, object]:
    return {k: v for k, v in x.items() if v is not None}


# Intended for use with json.load()'s object_hook parameter.
# Note also that object_hook is only called on dicts, not on lists, so this
# won't work for converting "top-level" lists to tuples
def lists_to_tuples(x: object) -> object:
    if isinstance(x, dict):
        x = cast(Dict[str, object], x)
        return {k: lists_to_tuples(v) for k, v in x.items()}
    elif isinstance(x, list):
        x = cast(List[object], x)
        return tuple(lists_to_tuples(y) for y in x)
    else:
        # TODO: are there other mutable iterators that we want to make read only?
        return x


def guess_mime_type(
    url: Union[str, "os.PathLike[str]"],
    default: str = "application/octet-stream",
    strict: bool = True,
) -> str:
    """
    Guess the MIME type of a file. This is a wrapper for mimetypes.guess_type, but it
    only returns the type (and not encoding), and it allows a default value.
    """
    # In Python<=3.7, mimetypes.guess_type only accepts strings.
    #
    # Note that in the parameters above, "os.PathLike[str]" is in quotes to avoid
    # "TypeError: 'ABCMeta' object is not subscriptable", in Python<=3.8.
    if sys.version_info < (3, 8):
        url = os.fspath(url)
    return mimetypes.guess_type(url, strict)[0] or default


# ==============================================================================
# Private random stream
# ==============================================================================
def private_random_int(min: int, max: int) -> str:
    with private_seed():
        return str(random.randint(min, max))


@contextlib.contextmanager
def private_seed():
    state = random.getstate()
    global own_random_state
    try:
        random.setstate(own_random_state)
        yield
    finally:
        own_random_state = random.getstate()
        random.setstate(state)


# Initialize random state for shiny's own private stream of randomness.
current_random_state = random.getstate()
random.seed(secrets.randbits(128))
own_random_state = random.getstate()
random.setstate(current_random_state)

# ==============================================================================
# Async-related functions
# ==============================================================================

T = TypeVar("T")


def wrap_async(
    fn: Union[Callable[[], T], Callable[[], Awaitable[T]]]
) -> Callable[[], Awaitable[T]]:
    """
    Given a synchronous function that returns T, return an async function that wraps the
    original function. If the input function is already async, then return it unchanged.
    """

    if is_async_callable(fn):
        return fn

    fn = cast(Callable[[], T], fn)

    @functools.wraps(fn)
    async def fn_async() -> T:
        return fn()

    return fn_async


def is_async_callable(
    obj: Union[Callable[..., T], Callable[..., Awaitable[T]]]
) -> TypeGuard[Callable[..., Awaitable[T]]]:
    """
    Returns True if `obj` is an `async def` function, or if it's an object with a
    `__call__` method which is an `async def` function. This function should generally
    be used in this code base instead of iscoroutinefunction().
    """
    if inspect.iscoroutinefunction(obj):
        return True
    if hasattr(obj, "__call__"):  # noqa: B004
        if inspect.iscoroutinefunction(obj.__call__):  # type: ignore
            return True

    return False


# See https://stackoverflow.com/a/59780868/412655 for an excellent explanation
# of how this stuff works.
# For a more in-depth explanation, see
# https://snarky.ca/how-the-heck-does-async-await-work-in-python-3-5/.
def run_coro_sync(coro: Awaitable[T]) -> T:
    """
    Run a coroutine that is in fact synchronous. Given a coroutine (which is
    returned by calling an `async def` function), this function will run the
    coroutine for one iteration. If the coroutine completes, then return the
    value. If it does not complete, then it will throw a `RuntimeError`.

    What it means to be "in fact synchronous": the coroutine must not yield
    control to the event loop. A coroutine may have an `await` expression in it,
    and that may call another function that has an `await`, but the chain will
    only yield control if a `yield` statement bubbles through `await`s all the
    way up. For example, `await asyncio.sleep(0)` will have a `yield` which
    bubbles up to the next level. Note that a `yield` in a generator used the
    regular way (not with `await`) will not bubble up, since it is not awaited
    on.
    """
    if not inspect.iscoroutine(coro):
        raise TypeError("run_coro_sync requires a Coroutine object.")

    try:
        coro.send(None)
    except StopIteration as e:
        return e.value

    raise RuntimeError(
        "async function yielded control; it did not finish in one iteration."
    )


def run_coro_hybrid(coro: Awaitable[T]) -> "asyncio.Future[T]":
    """
    Synchronously runs the given coro up to its first yield, then runs the rest of the
    coro by scheduling it on the current event loop, as per normal. You can think of
    this as either a run_coro_sync() that keeps running in the future, or, as an
    asyncio.create_task() that starts executing immediately instead of via call_soon.

    The status/result/exception can be access through the returned future. Even if an
    error happens synchronously, run_coro_hybrid() will not throw, but rather the error
    will be reported through the future object.

    **PLEASE ONLY USE THIS IF IT'S ABSOLUTELY NECESSARY.** Relative to the official
    asyncio Task implementation, this is a hastily assembled hack job; who knows what
    unknown unknowns lurk here.
    """
    result_future: asyncio.Future[T] = asyncio.Future()

    if not inspect.iscoroutine(coro):
        raise TypeError("run_coro_hybrid requires a Coroutine object.")

    # Inspired by Task.__step method in cpython/Lib/asyncio/tasks.py
    def _step(fut: Optional["asyncio.Future[None]"] = None):
        assert result_future.cancelled() or not result_future.done()

        exc: Optional[BaseException] = None
        if fut:
            assert fut.done()
            try:
                fut.result()
            except BaseException as e:
                exc = e

        if result_future.cancelled():
            # This may cause fut.result()'s exception to be ignored. That's intentional.
            # The cancellation takes precedent, but if we don't call fut.result() first
            # to retrieve its error, Python will warn.
            exc = CancelledError()

        res: Optional[asyncio.Future[None]] = None
        try:
            if exc is None:
                res = coro.send(None)
            else:
                # Is it worth throwing here? Or just logging?
                res = coro.throw(exc)
        except StopIteration as e:
            # Done
            result_future.set_result(e.value)
            return
        except CancelledError:
            result_future.cancel()
            return
        except (KeyboardInterrupt, SystemExit) as e:
            result_future.set_exception(e)
            raise
        except BaseException as e:
            result_future.set_exception(e)
        else:
            # If we get here, the coro didn't finish. Schedule it for completion.
            if isinstance(res, asyncio.Future):
                res.add_done_callback(_step)
            elif res is None:
                # This case happens with asyncio.sleep(0)
                asyncio.get_running_loop().call_soon(_step)
            else:
                raise RuntimeError(f"coroutine yielded unknown value: {res!r}")

    _step()

    return result_future


# ==============================================================================
# Callback registry
# ==============================================================================
class Callbacks:
    def __init__(self) -> None:
        self._callbacks: dict[int, Tuple[Callable[[], None], bool]] = {}
        self._id: int = 0

    def register(
        self, fn: Callable[[], None], once: bool = False
    ) -> Callable[[], None]:
        self._id += 1
        id = self._id
        self._callbacks[id] = (fn, once)

        def _():
            if id in self._callbacks:
                del self._callbacks[id]

        return _

    def invoke(self) -> None:
        # The list() wrapper is necessary to force collection of all the items before
        # iteration begins. This is necessary because self._callbacks may be mutated
        # by callbacks.
        for id, value in list(self._callbacks.items()):
            fn, once = value
            try:
                fn()
            finally:
                if once:
                    if id in self._callbacks:
                        del self._callbacks[id]

    def count(self) -> int:
        return len(self._callbacks)


class AsyncCallbacks:
    def __init__(self) -> None:
        self._callbacks: dict[int, Tuple[Callable[[], Awaitable[None]], bool]] = {}
        self._id: int = 0

    def register(
        self, fn: Callable[[], Awaitable[None]], once: bool = False
    ) -> Callable[[], None]:
        self._id += 1
        id = self._id
        self._callbacks[id] = (fn, once)

        def _():
            if id in self._callbacks:
                del self._callbacks[id]

        return _

    async def invoke(self) -> None:
        # The list() wrapper is necessary to force collection of all the items before
        # iteration begins. This is necessary because self._callbacks may be mutated
        # by callbacks.
        for id, value in list(self._callbacks.items()):
            fn, once = value
            try:
                await fn()
            finally:
                if once:
                    if id in self._callbacks:
                        del self._callbacks[id]

    def count(self) -> int:
        return len(self._callbacks)


# ==============================================================================
# System-related functions
# ==============================================================================

# Return directory that a package lives in.
def package_dir(package: str) -> str:
    with tempfile.TemporaryDirectory():
        pkg_file = importlib.import_module(".", package=package).__file__
        if pkg_file is None:
            raise RuntimeError(f"Could not find package dir for '{package}'")
        return os.path.dirname(pkg_file)



__all__ = ("resolve_id", "ui", "server")

import sys
from typing import Callable, TypeVar

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec, Concatenate
else:
    from typing import ParamSpec, Concatenate

from ._namespaces import resolve_id, namespace_context, Id
from .session import Inputs, Outputs, Session, require_active_session, session_context

P = ParamSpec("P")
R = TypeVar("R")


def ui(fn: Callable[P, R]) -> Callable[Concatenate[str, P], R]:
    def wrapper(id: Id, *args: P.args, **kwargs: P.kwargs) -> R:
        with namespace_context(id):
            return fn(*args, **kwargs)

    return wrapper


def server(
    fn: Callable[Concatenate[Inputs, Outputs, Session, P], R]
) -> Callable[Concatenate[str, P], R]:
    def wrapper(id: Id, *args: P.args, **kwargs: P.kwargs) -> R:
        sess = require_active_session(None)
        child_sess = sess.make_scope(id)
        with session_context(child_sess):
            return fn(child_sess.input, child_sess.output, child_sess, *args, **kwargs)

    return wrapper



import asyncio
from datetime import date

import numpy as np
from shiny import *

app_ui = ui.page_fluid(
    ui.download_button("downloadData", "Download"),
)

# For more examples of different types of download handlers, see:
# https://github.com/rstudio/py-shiny/blob/68ffc27/examples/download/app.py#L90
def server(input: Inputs, output: Outputs, session: Session):
    @session.download(
        filename=lambda: f"新型-{date.today().isoformat()}-{np.random.randint(100,999)}.csv"
    )
    async def downloadData():
        await asyncio.sleep(0.25)
        yield "one,two,three\n"
        yield "新,1,2\n"
        yield "型,4,5\n"


app = App(app_ui, server)
